# Copyright (c) Alibaba, Inc. and its affiliates.
import json
import os
from glob import glob

import numpy as np
import torch

from easycv.core.visualization import imshow_bboxes
from easycv.datasets.utils import replace_ImageToTensor
from easycv.file import io
from easycv.models.detection.utils import postprocess
from easycv.thirdparty.mtcnn import FaceDetector
from easycv.utils.checkpoint import load_checkpoint
from easycv.utils.misc import deprecated
from .base import InputProcessor, OutputProcessor, PredictorV2
from .builder import PREDICTORS
from .classifier import TorchClassifier

try:
    from easy_vision.python.inference.predictor import PredictorInterface
except Exception:
    from .interface import PredictorInterface


# 将张量转化为ndarray格式
def onnx_to_numpy(tensor):
    return tensor.detach().cpu().numpy(
    ) if tensor.requires_grad else tensor.cpu().numpy()


class DetInputProcessor(InputProcessor):

    def build_processor(self):
        if self.pipelines is not None:
            pipelines = self.pipelines
        elif self.cfg is None:
            pipelines = []
        else:
            pipelines = self.cfg.get('test_pipeline', [])

        # for batch inference
        self.pipelines = replace_ImageToTensor(pipelines)

        return super().build_processor()


class DetOutputProcessor(OutputProcessor):

    def __init__(self, score_thresh, classes=None):
        super(DetOutputProcessor, self).__init__()
        self.score_thresh = score_thresh
        self.classes = classes

    def process_single(self, inputs):
        if inputs['detection_scores'] is None or len(
                inputs['detection_scores']) < 1:
            return inputs

        scores = inputs['detection_scores']
        if scores is not None and self.score_thresh > 0:
            keeped_ids = scores > self.score_thresh
            inputs['detection_scores'] = inputs['detection_scores'][keeped_ids]
            inputs['detection_boxes'] = inputs['detection_boxes'][keeped_ids]
            inputs['detection_classes'] = inputs['detection_classes'][
                keeped_ids]

        class_names = []
        for _, classes_id in enumerate(inputs['detection_classes']):
            if classes_id is None:
                class_names.append(None)
            elif self.classes is not None and len(self.classes) > 0:
                class_names.append(self.classes[int(classes_id)])
            else:
                class_names.append(classes_id)

        inputs['detection_class_names'] = class_names

        return inputs


@PREDICTORS.register_module()
class DetectionPredictor(PredictorV2):
    """Generic Detection Predictor, it will filter bbox results by ``score_threshold`` .

    Args:
        model_path (str): Path of model path.
        config_file (Optinal[str]): config file path for model and processor to init. Defaults to None.
        batch_size (int): batch size for forward.
        device (str | torch.device): Support str('cuda' or 'cpu') or torch.device, if is None, detect device automatically.
        save_results (bool): Whether to save predict results.
        save_path (str): File path for saving results, only valid when `save_results` is True.
        pipelines (list[dict]): Data pipeline configs.
        input_processor_threads (int): Number of processes to process inputs.
        mode (str): The image mode into the model.
    """

    def __init__(self,
                 model_path,
                 config_file=None,
                 batch_size=1,
                 device=None,
                 save_results=False,
                 save_path=None,
                 pipelines=None,
                 score_threshold=0.5,
                 input_processor_threads=8,
                 mode='BGR',
                 *arg,
                 **kwargs):
        super(DetectionPredictor, self).__init__(
            model_path,
            config_file=config_file,
            batch_size=batch_size,
            device=device,
            save_results=save_results,
            save_path=save_path,
            pipelines=pipelines,
            input_processor_threads=input_processor_threads,
            mode=mode)
        self.score_thresh = score_threshold
        self.CLASSES = self.cfg.get('CLASSES', None)

    def get_input_processor(self):
        return DetInputProcessor(
            self.cfg,
            pipelines=self.pipelines,
            batch_size=self.batch_size,
            threads=self.input_processor_threads,
            mode=self.mode)

    def get_output_processor(self):
        return DetOutputProcessor(self.score_thresh, self.CLASSES)

    def visualize(self, img, results, show=False, out_file=None):
        """Only support show one sample now."""
        bboxes = results['detection_boxes']
        labels = results['detection_class_names']
        img = self.input_processor._load_input(img)['img']
        imshow_bboxes(
            img,
            bboxes,
            labels=labels,
            colors='cyan',
            text_color='cyan',
            font_size=18,
            thickness=2,
            font_scale=0.0,
            show=show,
            out_file=out_file)


class _JitProcessorWrapper:

    def __init__(self, processor, device) -> None:
        self.processor = processor
        self.device = device

    def __call__(self, results):
        if self.processor is not None:
            from mmcv.parallel import DataContainer as DC
            outputs = {}
            img = results['img']
            img = torch.from_numpy(img).to(self.device)
            img, img_meta = self.processor(img.unsqueeze(0))  # process batch
            outputs['img'] = DC(
                img.squeeze(0),
                stack=True)  # DC wrapper for collate batch and to device
            outputs['img_metas'] = DC(img_meta, cpu_only=True)
            return outputs
        return results


class YoloXInputProcessor(DetInputProcessor):
    """Input processor for yolox.

    Args:
        cfg (Config): Config instance.
        pipelines (list[dict]): Data pipeline configs.
        batch_size (int): batch size for forward.
        model_type (str): "raw" or "jit" or "blade"
        jit_processor_path (str): File of the saved processing operator of torch jit type.
        device (str | torch.device): Support str('cuda' or 'cpu') or torch.device, if is None, detect device automatically.
        threads (int): Number of processes to process inputs.
        mode (str): The image mode into the model.
    """

    def __init__(
        self,
        cfg,
        pipelines=None,
        batch_size=1,
        model_type='raw',
        jit_processor_path=None,
        device=None,
        threads=8,
        mode='BGR',
    ):
        self.model_type = model_type
        self.jit_processor_path = jit_processor_path
        self.device = device
        if self.device is None:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        super().__init__(
            cfg,
            pipelines=pipelines,
            batch_size=batch_size,
            threads=threads,
            mode=mode)

    def build_processor(self):
        self.jit_preprocess = False
        if self.model_type != 'raw':
            if hasattr(self.cfg, 'export'):
                self.jit_preprocess = self.cfg['export'].get(
                    'preprocess_jit', False)

        if self.model_type != 'raw' and self.jit_preprocess:
            # jit or blade model
            processor = None
            if os.path.exists(self.jit_processor_path):
                if self.threads > 1:
                    raise ValueError(
                        'Not support threads>1 for jit processor !')
                # use a preprocess jit model to speed up
                with io.open(self.jit_processor_path, 'rb') as infile:
                    processor = torch.jit.load(infile, self.device)
            return _JitProcessorWrapper(processor, self.device)
        else:
            return super().build_processor()


class YoloXOutputProcessor(DetOutputProcessor):

    def __init__(self,
                 score_thresh=0.5,
                 model_type='raw',
                 test_conf=0.01,
                 nms_thre=0.65,
                 use_trt_efficientnms=False,
                 classes=None):
        super().__init__(score_thresh, classes)
        self.model_type = model_type
        self.test_conf = test_conf
        self.nms_thre = nms_thre
        self.use_trt_efficientnms = use_trt_efficientnms

    def post_assign(self, outputs, img_metas):
        detection_boxes = []
        detection_scores = []
        detection_classes = []
        img_metas_list = []

        for i in range(len(outputs)):
            if img_metas:
                img_metas_list.append(img_metas[i])
            if outputs[i].requires_grad == True:
                outputs[i] = outputs[i].detach()

            if outputs[i] is not None:
                bboxes = outputs[i][:, 0:4] if outputs[i] is not None else None
                if img_metas:
                    bboxes /= img_metas[i]['scale_factor'][0]
                detection_boxes.append(bboxes.cpu().numpy())
                detection_scores.append(
                    (outputs[i][:, 4] * outputs[i][:, 5]).cpu().numpy())
                detection_classes.append(outputs[i][:, 6].cpu().numpy().astype(
                    np.int32))
            else:
                detection_boxes.append(None)
                detection_scores.append(None)
                detection_classes.append(None)

        test_outputs = {
            'detection_boxes': detection_boxes,
            'detection_scores': detection_scores,
            'detection_classes': detection_classes,
            'img_metas': img_metas_list
        }
        return test_outputs

    def process_single(self, inputs):
        det_out = inputs
        img_meta = det_out['img_metas']

        if self.model_type != 'raw':
            results = det_out['results']
            if self.use_trt_efficientnms:
                det_out = {}
                det_out['detection_boxes'] = results[1] / img_meta[
                    'scale_factor'][0]
                det_out['detection_scores'] = results[2]
                det_out['detection_classes'] = results[3]
            else:
                if self.model_type == 'jit':
                    det_out = self.post_assign(
                        results.unsqueeze(0), img_metas=[img_meta])
                else:
                    det_out = self.post_assign(
                        postprocess(
                            results.unsqueeze(0), len(self.classes),
                            self.test_conf, self.nms_thre),
                        img_metas=[img_meta])
            det_out['detection_scores'] = det_out['detection_scores'][0]
            det_out['detection_boxes'] = det_out['detection_boxes'][0]
            det_out['detection_classes'] = det_out['detection_classes'][0]

        resuts = super().process_single(det_out)
        resuts['ori_img_shape'] = list(img_meta['ori_img_shape'][:2])
        return resuts


@PREDICTORS.register_module()
class YoloXPredictor(DetectionPredictor):
    """Detection predictor for Yolox.

    Args:
        model_path (str): Path of model path.
        config_file (Optinal[str]): config file path for model and processor to init. Defaults to None.
        batch_size (int): batch size for forward.
        use_trt_efficientnms (bool): Whether used tensorrt efficient nms operation in the saved model.
        device (str | torch.device): Support str('cuda' or 'cpu') or torch.device, if is None, detect device automatically.
        save_results (bool): Whether to save predict results.
        save_path (str): File path for saving results, only valid when `save_results` is True.
        pipelines (list[dict]): Data pipeline configs.
        max_det (int): Maximum number of detection output boxes.
        score_thresh (float): Score threshold to filter box.
        nms_thresh (float): Nms threshold to filter box.
        input_processor_threads (int): Number of processes to process inputs.
        mode (str): The image mode into the model.
    """

    def __init__(self,
                 model_path,
                 config_file=None,
                 batch_size=1,
                 use_trt_efficientnms=False,
                 device=None,
                 save_results=False,
                 save_path=None,
                 pipelines=None,
                 max_det=100,
                 score_thresh=0.5,
                 nms_thresh=None,
                 test_conf=None,
                 input_processor_threads=8,
                 mode='BGR',
                 model_type=None):
        self.max_det = max_det
        self.use_trt_efficientnms = use_trt_efficientnms
        self.model_type = model_type
        if self.model_type is None:
            if model_path.endswith('jit'):
                self.model_type = 'jit'
            elif model_path.endswith('blade'):
                self.model_type = 'blade'
            elif model_path.endswith('onnx'):
                self.model_type = 'onnx'
            else:
                self.model_type = 'raw'
        assert self.model_type in ['raw', 'jit', 'blade', 'onnx']

        if self.model_type == 'blade' or self.use_trt_efficientnms:
            import torch_blade

        if self.model_type != 'raw' and config_file is None:
            config_file = model_path + '.config.json'

        super(YoloXPredictor, self).__init__(
            model_path,
            config_file=config_file,
            batch_size=batch_size,
            device=device,
            save_results=save_results,
            save_path=save_path,
            pipelines=pipelines,
            score_threshold=score_thresh,
            input_processor_threads=input_processor_threads,
            mode=mode)

        self.test_conf = test_conf or self.cfg['model'].get('test_conf', 0.01)
        self.nms_thre = nms_thresh or self.cfg['model'].get('nms_thre', 0.65)
        self.CLASSES = self.cfg.get('CLASSES', None) or self.cfg.get(
            'classes', None)
        assert self.CLASSES is not None
        self.jit_processor_path = '.'.join(
            self.model_path.split('.')[:-1] + ['preprocess'])

    def _build_model(self):
        if self.model_type != 'raw':
            if self.model_type != 'onnx':
                with io.open(self.model_path, 'rb') as infile:
                    model = torch.jit.load(infile, self.device)
            else:
                import onnxruntime
                if onnxruntime.get_device() == 'GPU':
                    model = onnxruntime.InferenceSession(
                        self.model_path, providers=['CUDAExecutionProvider'])
                else:
                    model = onnxruntime.InferenceSession(self.model_path)
        else:
            from easycv.utils.misc import reparameterize_models
            model = super()._build_model()
            model = reparameterize_models(model)
        return model

    def prepare_model(self):
        """Build model from config file by default.
        If the model is not loaded from a configuration file, e.g. torch jit model, you need to reimplement it.
        """
        model = self._build_model()
        if self.model_type != 'onnx':
            model.to(self.device)
            model.eval()
        if self.model_type == 'raw':
            load_checkpoint(model, self.model_path, map_location='cpu')
        return model

    def model_forward(self, inputs):
        """Model forward.
        If you need refactor model forward, you need to reimplement it.
        """
        if self.model_type != 'raw':
            with torch.no_grad():
                if self.model_type != 'onnx':
                    outputs = self.model(inputs['img'])
                else:
                    outputs = self.model.run(
                        None, {
                            self.model.get_inputs()[0].name:
                            onnx_to_numpy(inputs['img'])
                        })[0]
                    outputs = torch.from_numpy(outputs)
                outputs = {'results': outputs}  # convert to dict format
        else:
            outputs = super().model_forward(inputs)

        if 'img_metas' not in outputs:
            outputs['img_metas'] = inputs['img_metas']

        return outputs

    def get_input_processor(self):
        return YoloXInputProcessor(
            self.cfg,
            pipelines=self.pipelines,
            batch_size=self.batch_size,
            model_type=self.model_type,
            jit_processor_path=self.jit_processor_path,
            device=self.device,
            threads=self.input_processor_threads,
            mode=self.mode,
        )

    def get_output_processor(self):
        return YoloXOutputProcessor(
            score_thresh=self.score_thresh,
            model_type=self.model_type,
            test_conf=self.test_conf,
            nms_thre=self.nms_thre,
            use_trt_efficientnms=self.use_trt_efficientnms,
            classes=self.CLASSES)


@deprecated(reason='Please use YoloXPredictor.')
@PREDICTORS.register_module()
class TorchYoloXPredictor(YoloXPredictor):

    def __init__(self,
                 model_path,
                 max_det=100,
                 score_thresh=0.5,
                 use_trt_efficientnms=False,
                 model_config=None,
                 input_processor_threads=8,
                 mode='BGR'):
        """
        Args:
          model_path: model file path
          max_det: maximum number of detection
          score_thresh:  score_thresh to filter box
          model_config: config string for model to init, in json format
        """
        if model_config:
            if isinstance(model_config, str):
                model_config = json.loads(model_config)
        else:
            model_config = {}

        score_thresh = model_config[
            'score_thresh'] if 'score_thresh' in model_config else score_thresh
        super().__init__(
            model_path,
            config_file=None,
            batch_size=1,
            use_trt_efficientnms=use_trt_efficientnms,
            device=None,
            save_results=False,
            save_path=None,
            pipelines=None,
            max_det=max_det,
            score_thresh=score_thresh,
            nms_thresh=None,
            test_conf=None,
            input_processor_threads=input_processor_threads,
            mode=mode)

    def predict(self, input_data_list, batch_size=-1, to_numpy=True):
        return super().__call__(input_data_list)


@PREDICTORS.register_module()
class TorchFaceDetector(PredictorInterface):

    def __init__(self, model_path=None, model_config=None):
        """
    init model, add a facedetect and align for img input.

    Args:
      model_path: model file path
      model_config: config string for model to init, in json format
    """

        self.detector = FaceDetector()

    def get_output_type(self):
        """
    in this function user should return a type dict, which indicates
    which type of data should the output of predictor be converted to
    * type json, data will be serialized to json str

    * type image, data will be converted to encode image binary and write to oss file,
      whose name is output_dir/${key}/${input_filename}_${idx}.jpg, where input_filename
      is the base filename extracted from url, key corresponds to the key in the dict of output_type,
      if the type of data indexed by key is a list, idx is the index of element in list, otherwhile ${idx} will be empty

    * type video, data will be converted to encode video binary and write to oss file,

    :: return  {
      'image': 'image',
      'feature': 'json'
    }
    indicating that the image data in the output dict will be save to image
    file and feature in output dict will be converted to json

    """
        return {}

    def batch(self, image_tensor_list):
        return torch.stack(image_tensor_list)

    def predict(self, input_data_list, batch_size=-1, threshold=0.95):
        """
    using session run predict a number of samples using batch_size

    Args:
      input_data_list:  a list of numpy array, each array is a sample to be predicted
      batch_size: batch_size passed by the caller, you can also ignore this param and
        use a fixed number if you do not want to adjust batch_size in runtime
    Return:
      result: a list of dict, each dict is the prediction result of one sample
        eg, {"output1": value1, "output2": value2}, the value type can be
        python int str float, and numpy array
    Raise:
       if detect !=1 face in a img, then do nothing for this image
    """
        num_image = len(input_data_list)
        assert len(
            input_data_list) > 0, 'input images should not be an empty list'

        image_list = input_data_list
        outputs_list = []

        for idx, img in enumerate(image_list):
            if type(img) is not np.ndarray:
                img = np.asarray(img)

            ori_img_shape = img.shape[:2]
            bbox, ld = self.detector.safe_detect(img)
            _scores = np.array([i[-1] for i in bbox])

            boxes = []
            scores = []
            for idx, s in enumerate(_scores):
                if s > threshold:
                    boxes.append(bbox[idx][:-1])
                    scores.append(bbox[idx][-1])
            boxes = np.array(boxes)
            scores = np.array(scores)

            out = {
                'ori_img_shape': list(ori_img_shape),
                'detection_boxes': boxes,
                'detection_scores': scores,
                'detection_classes': [0] * boxes.shape[0],
                'detection_class_names': ['face'] * boxes.shape[0],
            }

            outputs_list.append(out)

        return outputs_list


@PREDICTORS.register_module()
class TorchYoloXClassifierPredictor(PredictorInterface):

    def __init__(self,
                 models_root_dir,
                 max_det=100,
                 cls_score_thresh=0.01,
                 det_model_config=None,
                 cls_model_config=None):
        """
    init model, add a yolox and classification predictor for img input.

    Args:
      models_root_dir: models_root_dir/detection/*.pth and models_root_dir/classification/*.pth
      det_model_config: config string for detection model to init, in json format
      cls_model_config: config string for classification model to init, in json format
    """
        det_model_path = glob(
            '%s/detection/*.pt*' % models_root_dir, recursive=True)
        assert (len(det_model_path) == 1)
        cls_model_path = glob(
            '%s/classification/*.pt*' % models_root_dir, recursive=True)
        assert (len(cls_model_path) == 1)

        self.det_predictor = TorchYoloXPredictor(
            det_model_path[0], max_det=max_det, model_config=det_model_config)
        self.cls_predictor = TorchClassifier(
            cls_model_path[0], model_config=cls_model_config)
        self.cls_score_thresh = cls_score_thresh

    def predict(self, input_data_list, batch_size=-1):
        """
    using session run predict a number of samples using batch_size

    Args:
      input_data_list:  a list of numpy array(in rgb order), each array is a sample
        to be predicted
      batch_size: batch_size passed by the caller, you can also ignore this param and
        use a fixed number if you do not want to adjust batch_size in runtime
    Return:
      result: a list of dict, each dict is the prediction result of one sample
        eg, {"output1": value1, "output2": value2}, the value type can be
        python int str float, and numpy array
    """
        results = self.det_predictor.predict(
            input_data_list, batch_size=batch_size)

        for img_idx, img in enumerate(input_data_list):
            detection_boxes = results[img_idx]['detection_boxes']
            detection_classes = results[img_idx]['detection_classes']
            detection_scores = results[img_idx]['detection_scores']

            crop_img_batch = []
            for idx in range(detection_boxes.shape[0]):
                xyxy = [int(a) for a in detection_boxes[idx]]
                cropImg = img[xyxy[1]:xyxy[3], xyxy[0]:xyxy[2]]
                crop_img_batch.append(cropImg)

            if len(crop_img_batch) > 0:
                cls_output = self.cls_predictor.predict(
                    crop_img_batch, batch_size=32)
            else:
                cls_output = []

            class_name_list = []
            class_id_list = []
            class_score_list = []
            det_bboxes = []
            product_count_dict = {}

            for idx in range(len(cls_output)):
                class_name = cls_output[idx]['class_name'][0]
                class_score = cls_output[idx]['class_probs'][class_name]
                if class_score < self.cls_score_thresh:
                    continue

                if class_name not in product_count_dict:
                    product_count_dict[class_name] = 1
                else:
                    product_count_dict[class_name] += 1
                class_name_list.append(class_name)
                class_id_list.append(int(cls_output[idx]['class'][0]))
                class_score_list.append(class_score)
                det_bboxes.append([float(a) for a in detection_boxes[idx]])

            results[img_idx].update({
                'detection_boxes': np.array(det_bboxes),
                'detection_scores': class_score_list,
                'detection_classes': class_id_list,
                'detection_class_names': class_name_list,
                'product_count': product_count_dict
            })

        return results



# xuebinqin/U-2-Net is licensed under the

# Apache License 2.0
# A permissive license whose main conditions require preservation of copyright and license notices. Contributors provide an express grant of patent rights. Licensed works, modifications, and larger works may be distributed under different terms and without source code.

# Permissions
#  Commercial use
#  Modification
#  Distribution
#  Patent use
#  Private use


"""
Mostly copy-paste from
https://github.com/facebookresearch/mae/blob/main/models_mae.py
"""

from functools import partial

import torch
import torch.nn as nn
from timm.models.vision_transformer import Block, PatchEmbed

from easycv.models.utils import get_2d_sincos_pos_embed
from ..registry import BACKBONES


@BACKBONES.register_module
class MaskedAutoencoderViT(nn.Module):
    """ Masked Autoencoder with VisionTransformer backbone.
        MaskedAutoencoderViT is mostly same as vit_tranformer_dynamic, but with a random_masking func.
        MaskedAutoencoderViT model can be loaded by vit_tranformer_dynamic.

    Args:
        img_size(int): input image size
        patch_size(int): patch size
        in_chans(int): input image channels
        embed_dim(int): feature dimensions
        depth(int): number of encoder layers
        num_heads(int): Parallel attention heads
        mlp_ratio(float): mlp ratio
        norm_layer: type of normalization layer
    """

    def __init__(
            self,
            img_size=224,
            patch_size=16,
            in_chans=3,
            embed_dim=1024,
            depth=24,
            num_heads=16,
            mlp_ratio=4.,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
    ):
        super().__init__()

        self.patch_size = patch_size
        self.patch_embed = PatchEmbed(img_size, patch_size, in_chans,
                                      embed_dim)
        self.num_patches = self.patch_embed.num_patches
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(
            torch.zeros(1, self.num_patches + 1, embed_dim),
            requires_grad=False)  # fixed sin-cos embedding
        self.blocks = nn.ModuleList([
            Block(
                embed_dim,
                num_heads,
                mlp_ratio,
                qkv_bias=True,
                norm_layer=norm_layer) for i in range(depth)
        ])
        self.norm = norm_layer(embed_dim)

    def init_weights(self):
        w = self.patch_embed.proj.weight.data
        torch.nn.init.xavier_uniform_(w.view([w.shape[0], -1]))
        torch.nn.init.normal_(self.cls_token, std=.02)
        pos_embed = get_2d_sincos_pos_embed(
            self.pos_embed.shape[-1],
            int(self.patch_embed.num_patches**.5),
            cls_token=True)
        self.pos_embed.data.copy_(
            torch.from_numpy(pos_embed).float().unsqueeze(0))

        for m in self.modules():
            if isinstance(m, nn.Linear):
                # we use xavier_uniform following official JAX ViT:
                torch.nn.init.xavier_uniform_(m.weight)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

    def random_masking(self, x, mask_ratio):
        """
        Perform per-sample random masking by per-sample shuffling.
        Per-sample shuffling is done by argsort random noise.
        x: [N, L, D], sequence
        """
        N, L, D = x.shape  # batch, length, dim
        len_keep = int(L * (1 - mask_ratio))

        noise = torch.rand(N, L, device=x.device)  # noise in [0, 1]

        # sort noise for each sample
        ids_shuffle = torch.argsort(
            noise, dim=1)  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        x_masked = torch.gather(
            x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, D))

        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([N, L], device=x.device)
        mask[:, :len_keep] = 0
        # unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return x_masked, mask, ids_restore

    def forward(self, x, mask_ratio):
        # embed patches
        x = self.patch_embed(x)

        # add pos embed w/o cls token
        x = x + self.pos_embed[:, 1:, :]

        # masking: length -> length * mask_ratio
        x, mask, ids_restore = self.random_masking(x, mask_ratio)

        # append cls token
        cls_token = self.cls_token + self.pos_embed[:, :1, :]
        cls_tokens = cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # apply Transformer blocks
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)

        return x, mask, ids_restore



# model settings
model = dict(
    type='Detection',
    pretrained=True,
    backbone=dict(
        type='ResNet',
        depth=50,
        num_stages=4,
        out_indices=(2, 3, 4),
        frozen_stages=1,
        norm_cfg=dict(type='BN', requires_grad=False),
        norm_eval=True,
        style='pytorch'),
    head=dict(
        type='DINOHead',
        transformer=dict(
            type='DeformableTransformer',
            d_model=256,
            nhead=8,
            num_queries=900,
            num_encoder_layers=6,
            num_unicoder_layers=0,
            num_decoder_layers=6,
            dim_feedforward=2048,
            dropout=0.0,
            activation='relu',
            normalize_before=False,
            return_intermediate_dec=True,
            query_dim=4,
            num_patterns=0,
            modulate_hw_attn=True,
            # for deformable encoder
            deformable_encoder=True,
            deformable_decoder=True,
            num_feature_levels=4,
            enc_n_points=4,
            dec_n_points=4,
            # init query
            decoder_query_perturber=None,
            add_channel_attention=False,
            random_refpoints_xy=False,
            # two stage
            two_stage_type=
            'standard',  # ['no', 'standard', 'early', 'combine', 'enceachlayer', 'enclayer1']
            two_stage_pat_embed=0,
            two_stage_add_query_num=0,
            two_stage_learn_wh=False,
            two_stage_keep_all_tokens=False,
            # evo of #anchors
            dec_layer_number=None,
            rm_dec_query_scale=True,
            rm_self_attn_layers=None,
            key_aware_type=None,
            # layer share
            layer_share_type=None,
            # for detach
            rm_detach=None,
            decoder_sa_type='sa',
            module_seq=['sa', 'ca', 'ffn'],
            # for dn
            embed_init_tgt=True,
            use_detached_boxes_dec_out=False),
        dn_components=dict(
            dn_number=100,
            dn_label_noise_ratio=0.5,  # paper 0.5, release code 0.25
            dn_box_noise_scale=1.0,
            dn_labelbook_size=80,
        ),
        num_classes=80,
        in_channels=[512, 1024, 2048],
        embed_dims=256,
        query_dim=4,
        num_queries=900,
        num_select=300,
        random_refpoints_xy=False,
        num_patterns=0,
        fix_refpoints_hw=-1,
        num_feature_levels=4,
        # two stage
        two_stage_type='standard',  # ['no', 'standard']
        two_stage_add_query_num=0,
        dec_pred_class_embed_share=True,
        dec_pred_bbox_embed_share=True,
        two_stage_class_embed_share=False,
        two_stage_bbox_embed_share=False,
        decoder_sa_type='sa',
        temperatureH=20,
        temperatureW=20,
        cost_dict=dict(
            cost_class=2,
            cost_bbox=5,
            cost_giou=2,
        ),
        weight_dict=dict(loss_ce=1, loss_bbox=5, loss_giou=2)))



# Copyright (c) Alibaba, Inc. and its affiliates.
import logging
import os
import random

from PIL import Image, ImageFile

from easycv.datasets.registry import DATASOURCES
from easycv.file import io
from easycv.file.image import load_image
from easycv.utils.dist_utils import dist_zero_exec
from .utils import split_listfile_byrank


@DATASOURCES.register_module
class ClsSourceImageListByClass(object):
    """
    Get the same `m_per_class` samples by the label idx.

    Args:
        list_file : str / list(str), str means a input image list file path,
            this file contains records as  `image_path label` in list_file
            list(str) means multi image list, each one contains some records as `image_path label`
        root: str / list(str), root path for image_path, each list_file will need a root.
        m_per_class: num of samples for each class.
        delimeter: str, delimeter of each line in the `list_file`
        split_huge_listfile_byrank: Adapt to the situation that the memory cannot fully load a huge amount of data list.
            If split, data list will be split to each rank.
        cache_path: if `split_huge_listfile_byrank` is true, cache list_file will be saved to cache_path.
        max_try: int, max try numbers of reading image
    """

    def __init__(self,
                 root,
                 list_file,
                 m_per_class=2,
                 delimeter=' ',
                 split_huge_listfile_byrank=False,
                 cache_path='data/',
                 max_try=20):

        ImageFile.LOAD_TRUNCATED_IMAGES = True

        # TODO: support return list, donot save split file
        # TODO: support loading list_file that have already been split
        if split_huge_listfile_byrank:
            with dist_zero_exec():
                list_file = split_listfile_byrank(
                    list_file=list_file,
                    label_balance=True,
                    save_path=cache_path)

        with io.open(list_file, 'r') as f:
            lines = f.readlines()

        self.m_per_class = m_per_class
        self.has_labels = len(lines[0].split(delimeter)) >= 2
        assert self.has_labels is True

        label2files = {}

        for l in lines:
            label = int(l.strip().split(delimeter)[1])
            path = l.strip().split(delimeter)[0]

            if label in label2files.keys():
                label2files[label].append(path)
            else:
                label2files[label] = [path]

        self.labels = list(label2files.keys())
        self.fns_by_labels = [label2files[i] for i in self.labels]
        self.root = root

        self.initialized = False
        self.max_try = max_try

    def __len__(self):
        return len(self.fns_by_labels)

    def __getitem__(self, idx):
        label = self.labels[idx]
        image_list = self.fns_by_labels[idx]
        if len(image_list) < 1:
            logging.info('%s :image list contain < 1 image' % idx)
            return self[idx + 1]

        if self.m_per_class > len(image_list):
            image_list = int(self.m_per_class / len(image_list) +
                             1) * image_list

        sample_list = random.sample(image_list, self.m_per_class)
        return_img = []
        return_label = []

        for path in sample_list:
            img = load_image(os.path.join(self.root, path), mode='RGB')
            if img is None:
                return self[idx + 1]

            img = Image.fromarray(img)
            return_img.append(img)
            return_label.append(label)

        result_dict = {'img': return_img, 'gt_labels': return_label}
        return result_dict



# Copyright (c) Alibaba, Inc. and its affiliates.
import math
import random

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


def bninceptionPre(image, mean=[104, 117, 128], std=[1, 1, 1]):
    """
    Args:
        image: pytorch Image tensor from PIL (range 0~1), bgr format
        mean : norm mean
        std  : norm val
    Returns:
        A image norm in 0~255, rgb format
    """
    expand_batch_dim = len(image.size()) == 3
    if expand_batch_dim:
        image = image.unsqueeze(0)
    image = image * 255
    image = image[:, [2, 1, 0]]
    for i in range(2):
        image[:, i, ...] -= mean[i]
        image[:, i, ...] /= std[i]
    return image


def randomErasing(image,
                  probability=0.5,
                  sl=0.02,
                  sh=0.2,
                  r1=0.3,
                  mean=[0.4914, 0.4822, 0.4465]):
    expand_batch_dim = len(image.size()) == 3
    if expand_batch_dim:
        image = image.unsqueeze(0)

    batch_size = image.size(0)
    width = image.size(2)
    height = image.size(3)
    area = width * height

    for index in range(batch_size):
        erase_flag = False

        while not erase_flag:

            target_area = random.uniform(sl, sh) * area
            aspect_ratio = random.uniform(r1, 1 / r1)
            h = int(round(math.sqrt(target_area * aspect_ratio)))
            w = int(round(math.sqrt(target_area / aspect_ratio)))

            if w < width and h < height:
                x1 = random.randint(0, height - h)
                y1 = random.randint(0, width - w)
                image[index, 0, x1:x1 + h, y1:y1 + w] = mean[0]
                image[index, 1, x1:x1 + h, y1:y1 + w] = mean[1]
                image[index, 2, x1:x1 + h, y1:y1 + w] = mean[2]

                erase_flag = True

    return image


def solarize(tensor, threshold=0.5, apply_prob=0.2):
    """
        tensor : pytorch tensor
    """
    inverted_tensor = threshold * 2 - tensor
    mask = tensor >= threshold

    t = torch.ones_like(mask).type(torch.bool)  # a true mask
    return mask * inverted_tensor + (t ^ mask) * tensor


def gaussianBlurDynamic(image, apply_prob=0.5):
    expand_batch_dim = len(image.size()) == 3
    if expand_batch_dim:
        image = image.unsqueeze(0)

    batch_size = image.size(0)
    result = torch.zeros_like(image, device=image.device)

    for index in range(batch_size):
        if random.random() < apply_prob:
            sigma = random.uniform(0.1, 2.0)
            kernel_size = int(sigma * 4 + 0.5)
            radius = int(kernel_size / 2)
            kernel_size = radius * 2 + 1

            # x = torch.arange(-radius, radius + 1).cuda()
            x = torch.arange(-radius, radius + 1, device=image.device)
            x = x.to(image.dtype)
            blur_filter = torch.exp(-torch.pow(x, 2.0) / (2.0 * (sigma**2)))
            blur_filter = blur_filter.div(blur_filter.sum())
            blur_v = torch.reshape(blur_filter, [1, 1, kernel_size, 1])
            blur_h = torch.reshape(blur_filter, [1, 1, 1, kernel_size])

            num_channels, _, _ = image.size(1), image.size(2), image.size(3)
            blur_h = blur_h.repeat(num_channels, 1, 1, 1)
            blur_v = blur_v.repeat(num_channels, 1, 1, 1)
            pad_length = int((kernel_size - 1) / 2)
            blurred = F.conv2d(
                image[index:index + 1],
                blur_h,
                stride=1,
                padding=(0, pad_length),
                groups=3)
            blurred = F.conv2d(
                blurred, blur_v, stride=1, padding=(pad_length, 0), groups=3)
            if expand_batch_dim:
                blurred = blurred.squeeze(0)
            result[index] = blurred
        else:
            result[index] = image[index]

    return result


def gaussianBlur(image, kernel_size=22, apply_prob=0.5):
    expand_batch_dim = len(image.size()) == 3
    if expand_batch_dim:
        image = image.unsqueeze(0)

    batch_size = image.size(0)
    result = torch.zeros_like(image, device=image.device)

    for index in range(batch_size):
        if random.random() < apply_prob:
            sigma = random.uniform(0.1, 2.0)
            radius = int(kernel_size / 2)
            kernel_size = radius * 2 + 1
            # x = torch.arange(-radius, radius + 1).cuda()
            x = torch.arange(-radius, radius + 1, device=image.device)
            x = x.to(image.dtype)
            blur_filter = torch.exp(-torch.pow(x, 2.0) / (2.0 * (sigma**2)))
            blur_filter = blur_filter.div(blur_filter.sum())
            blur_v = torch.reshape(blur_filter, [1, 1, kernel_size, 1])
            blur_h = torch.reshape(blur_filter, [1, 1, 1, kernel_size])

            num_channels, _, _ = image.size(1), image.size(2), image.size(3)
            blur_h = blur_h.repeat(num_channels, 1, 1, 1)
            blur_v = blur_v.repeat(num_channels, 1, 1, 1)
            pad_length = int((kernel_size - 1) / 2)
            blurred = F.conv2d(
                image[index:index + 1],
                blur_h,
                stride=1,
                padding=(0, pad_length),
                groups=3)
            blurred = F.conv2d(
                blurred, blur_v, stride=1, padding=(pad_length, 0), groups=3)
            if expand_batch_dim:
                blurred = blurred.squeeze(0)
            result[index] = blurred
        else:
            result[index] = image[index]

    return result


def randomGrayScale(image, apply_prob=0.2):
    expand_batch_dim = len(image.size()) == 3
    if expand_batch_dim:
        image = image.unsqueeze(0)
    batch_size = image.size(0)
    for index in range(batch_size):
        if random.random() < apply_prob:
            tmp = 0.299 * image[index, 0] + 0.587 * image[
                index, 1] + 0.114 * image[index, 2]
            image[index, 0] = tmp
            image[index, 1] = tmp
            image[index, 2] = tmp
    return image


def mixUp(image, alpha=0.2):
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = image.size()[0]

    assert batch_size > 2
    assert batch_size % 2 == 0

    mixed_x = lam * image[0:int(batch_size / 2),
                          ...] + (1 - lam) * image[int(batch_size / 2):, ...]

    return mixed_x, lam


def mixUpCls(data, alpha=0.2):
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = data.size(0)
    index = torch.randperm(batch_size, device=data.device)
    data_mixed = lam * data + (1 - lam) * data[index, :]

    return data_mixed, lam, index


if __name__ == '__main__':
    a = torch.ones([12, 3, 224, 224])
    print(torch.sum(a))
    f = eval('gaussianBlur')
    print(type(f))
    b = f(a)
    print(torch.sum(b))

    print(b.shape)



# Copyright (c) Alibaba, Inc. and its affiliates.
import os

from torchvision.datasets import Caltech101, Caltech256
from torchvision.datasets.utils import (download_and_extract_archive,
                                        extract_archive)

from easycv.datasets.registry import DATASOURCES


@DATASOURCES.register_module
class ClsSourceCaltech101(object):

    def __init__(self, root, download=True):

        if download:
            root = self.download(root)
            self.caltech101 = Caltech101(root, 'category', download=False)
        else:
            self.caltech101 = Caltech101(root, 'category', download=False)

        # data label_classes
        self.CLASSES = self.caltech101.categories

    def __len__(self):
        return len(self.caltech101.index)

    def __getitem__(self, idx):
        # img: HWC, RGB
        img, label = self.caltech101[idx]
        print(self.caltech101[idx])
        result_dict = {'img': img, 'gt_labels': label}
        return result_dict

    def download(self, root):

        if os.path.exists(os.path.join(root, 'caltech101')):
            return root

        if os.path.exists(os.path.join(root, 'caltech-101.zip')):
            self.downloaded_exists(root)
            return root

        # download and extract the file
        download_and_extract_archive(
            'https://data.caltech.edu/records/mzrjq-6wc02/files/caltech-101.zip?download=1',
            root,
            filename='caltech-101.zip',
            md5='3138e1922a9193bfa496528edbbc45d0',
            remove_finished=True)
        self.normalized_path(root)

        return root

    # The data has been downloaded and decompressed
    def downloaded_exists(self, root):
        extract_archive(
            os.path.join(root, 'caltech-101.zip'), root, remove_finished=True)
        self.normalized_path(root)

    # The routinized path meets the input requirements
    def normalized_path(self, root):
        # rename root path
        old_folder_name = os.path.join(root, 'caltech-101')
        new_folder_name = os.path.join(root, 'caltech101')
        os.rename(old_folder_name, new_folder_name)
        # extract object file
        img_categories = os.path.join(new_folder_name,
                                      '101_ObjectCategories.tar.gz')
        extract_archive(img_categories, new_folder_name, remove_finished=True)


@DATASOURCES.register_module
class ClsSourceCaltech256(object):

    def __init__(self, root, download=True):

        if download:
            self.download(root)
            self.caltech256 = Caltech256(root, download=False)
        else:
            self.caltech256 = Caltech256(root, download=False)

        # data classes
        self.CLASSES = self.caltech256.categories

    def __len__(self):
        return len(self.caltech256.index)

    def __getitem__(self, idx):
        # img: HWC, RGB
        img, label = self.caltech256[idx]
        result_dict = {'img': img, 'gt_labels': label}
        return result_dict

    def download(self, root):

        caltech256_path = os.path.join(root, 'caltech256')

        if os.path.exists(caltech256_path):
            return

        download_and_extract_archive(
            'https://data.caltech.edu/records/nyy15-4j048/files/256_ObjectCategories.tar?download=1',
            caltech256_path,
            filename='256_ObjectCategories.tar',
            md5='67b4f42ca05d46448c6bb8ecd2220f6d',
            remove_finished=True)



# Copyright 2019 Alibaba Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Loss modules."""

from __future__ import absolute_import, division, print_function

import torch

from easycv.core.sailfish.activation import LogSoftmax
from easycv.core.sailfish.linear import ArcFaceLinear, Linear
from easycv.core.sailfish.util import ModelParallel


class NLLLoss(torch.nn.Module):
    r"""The negative log likelihood loss for log probabilities. It is
  useful to train a classification problem with `C` classes.

  The `input` given through a forward call is expected to contain
  log-probabilities of each class. `input` has to be a Tensor of size either
  :math:`(N, C)` or :math:`(N, C, d_1, d_2, ..., d_K)`
  with :math:`K \geq 1` for the `K`-dimensional case (described later).

  Obtaining log-probabilities in a neural network is easily achieved by
  adding a  `LogSoftmax` layer in the last layer of your network.
  You may use `CrossEntropyLoss` instead, if you prefer not to add an
  extra layer.

  The `target` that this loss expects should be a class index in the range
  :math:`[0, C-1]` where `C = number\_classes`.

  NLLLoss is defined as:
  .. math::
      \ell(x, y) = -\frac{1}{N}\sum_{n=1}^N L_{i}

  Args:
    num_classes: total number of classes.
    focal: whether to use FocalLoss implementation.
    focal_gamm: The focusing parameter of FocalLoss.
    rank: rank of current replica.
    world_size: size of replicas.

  Shape:
    - Input: :math:`(\frac{N}{P}, C)` where `C = number of classes`, or
      :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1`
      in the case of `K`-dimensional loss.
    - Target: :math:`(\frac{N}{P}, 1)` where each value is
      :math:`0 \leq \text{targets}[i] \leq C-1`, or
      :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
      K-dimensional loss.
    - Output: scalar.
      If :attr:`reduction` is ``'none'``, then the same size as the target:
      :math:`(N)`, or :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in
      the case of K-dimensional loss.
  Examples::
    >>> m = LogSoftmax(...)
    >>> loss = NLLLoss(...)
    >>> # input is of size N x C = 3 x 5
    >>> input = torch.randn(3, 5, requires_grad=True)
    >>> # each element in target has to have 0 <= value < C
    >>> target = torch.tensor([1, 0, 4])
    >>> output = loss(m(input), target)
    >>> output.backward()
  """

    def __init__(self, focal=False, focal_gamma=0, parallel=None):
        super(NLLLoss, self).__init__()
        self.focal = focal
        self.focal_gamma = focal_gamma
        self.parallel = parallel

    def forward(self, logprob, target):  # pylint: disable=arguments-differ
        """Compute negative log likelihood loss from log-probs and target."""
        if isinstance(self.parallel, ModelParallel):
            with torch.no_grad():
                mask = self.parallel.correct_mask(target, logprob)
            loss = self.parallel.nll_loss(logprob, mask)
            if self.focal:
                loss_exp = torch.exp(-loss)
                loss = (1 - loss_exp)**self.focal_gamma * loss
            return loss
        loss = torch.nn.functional.nll_loss(logprob, target)
        if self.focal:
            loss_exp = torch.exp(-loss)
            loss = (1 - loss_exp)**self.focal_gamma * loss
        return loss


class CrossEntropyLoss(torch.nn.Module):
    r"""This criterion combines :func:`LogSoftmax` and
    :func:`NLLLoss` in one single class.
  """

    def __init__(self, epsilon=0, parallel=None):
        super(CrossEntropyLoss, self).__init__()
        self._log_softmax = LogSoftmax(epsilon=epsilon, parallel=parallel)
        self._nll_loss = NLLLoss(parallel=parallel)

    def forward(self, logits, target):  # pylint: disable=arguments-differ
        # 1. Compute log-probabilities for current shard:
        logprob = self._log_softmax(logits)

        # 2. Compute NLL loss for all shards.
        return self._nll_loss(logprob, target)


class FocalLoss(torch.nn.Module):
    r"""This criterion combines :func:`LogSoftmax` and
    :func:`NLLLoss` in one single class.
  """

    def __init__(self, gamma=0, epsilon=1e-8, parallel=None):
        super(FocalLoss, self).__init__()
        self._log_softmax = LogSoftmax(epsilon=epsilon, parallel=parallel)
        self._nll_loss = NLLLoss(
            focal=True, focal_gamma=gamma, parallel=parallel)

    def forward(self, logits, target):  # pylint: disable=arguments-differ
        logprob = self._log_softmax(logits)
        return self._nll_loss(logprob, target)


class SoftmaxLoss(torch.nn.Module):
    r"""This criterion combines :func:`Linear` and
    :func:`CrossEntropyLoss` in one single class.
  """

    def __init__(self,
                 in_features,
                 out_features,
                 bias=True,
                 epsilon=0,
                 weight_initializer=None,
                 bias_initializer=None,
                 parallel=None):
        super(SoftmaxLoss, self).__init__()
        self._linear = Linear(
            in_features,
            out_features,
            bias=bias,
            weight_initializer=weight_initializer,
            bias_initializer=bias_initializer,
            parallel=parallel)
        self._log_softmax = LogSoftmax(epsilon=epsilon, parallel=parallel)
        self._nll_loss = NLLLoss(parallel=parallel)
        self._parallel = parallel

    def forward(self, features, target):  # pylint: disable=arguments-differ
        if isinstance(self._parallel, ModelParallel):
            features = self._parallel.gather(features)
        logits = self._linear(features.squeeze())
        logprob = self._log_softmax(logits)
        if isinstance(self._parallel, ModelParallel):
            target = self._parallel.gather_target(target)
        return self._nll_loss(logprob, target)


class ArcMarginLoss(torch.nn.Module):
    r"""This criterion combines :func:`ArcFaceLinear` and
    :func:`CrossEntropyLoss` in one single class.
  """

    def __init__(
            self,
            in_features,
            out_features,
            margin=0.5,
            scale=64.0,  # See normface https://arxiv.org/abs/1704.06369
            fast_phi=False,
            epsilon=0,
            weight_initializer=None,
            l2_norm=False,
            parallel=None):
        super(ArcMarginLoss, self).__init__()
        self._linear = ArcFaceLinear(
            in_features,
            out_features,
            margin=margin,
            scale=scale,
            l2_norm=l2_norm,
            fast_phi=fast_phi,
            epsilon=epsilon,
            weight_initializer=weight_initializer,
            parallel=parallel)
        self._log_softmax = LogSoftmax(epsilon=epsilon, parallel=parallel)
        self._nll_loss = NLLLoss(parallel=parallel)
        self._parallel = parallel

    def forward(self, features, target):  # pylint: disable=arguments-differ
        if isinstance(self._parallel, ModelParallel):
            features = self._parallel.gather(features)
            target = self._parallel.gather_target(target)
        logits = self._linear(features.squeeze(), target)
        logprob = self._log_softmax(logits)
        return self._nll_loss(logprob, target)



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) OpenMMLab. All rights reserved.
# Refer to: https://github.com/open-mmlab/mmaction2/blob/master/mmaction/models/skeleton_gcn/skeletongcn.py
from easycv.models.builder import MODELS
from .base import BaseGCN


@MODELS.register_module()
class SkeletonGCN(BaseGCN):
    """Spatial temporal graph convolutional networks."""

    def forward_train(self, skeletons, labels, **kwargs):
        """Defines the computation performed at every call when training."""
        assert self.with_cls_head
        losses = dict()

        x = self.extract_feat(skeletons)
        output = self.cls_head(x)
        gt_labels = labels.squeeze(-1)
        loss = self.cls_head.loss(output, gt_labels)
        losses.update(loss)

        return losses

    def forward_test(self, skeletons, **kwargs):
        """Defines the computation performed at every call when evaluation and
        testing."""
        x = self.extract_feat(skeletons)
        assert self.with_cls_head
        output = self.cls_head(x)

        result = {'prob': output.cpu()}
        return result



# Copyright (c) OpenMMLab. All rights reserved.
# This script converts MOT labels into COCO style.
# Official website of the MOT dataset: https://motchallenge.net/
#
# Label format of MOT dataset:
#   GTs:
#       <frame_id> # starts from 1 but COCO style starts from 0,
#       <instance_id>, <x1>, <y1>, <w>, <h>,
#       <conf> # conf is annotated as 0 if the object is ignored,
#       <class_id>, <visibility>
#
#   DETs and Results:
#       <frame_id>, <instance_id>, <x1>, <y1>, <w>, <h>, <conf>,
#       <x>, <y>, <z> # for 3D objects
#
# Classes in MOT:
#   1: 'pedestrian'
#   2: 'person on vehicle'
#   3: 'car'
#   4: 'bicycle'
#   5: 'motorbike'
#   6: 'non motorized vehicle'
#   7: 'static person'
#   8: 'distractor'
#   9: 'occluder'
#   10: 'occluder on the ground',
#   11: 'occluder full'
#   12: 'reflection'
#
#   USELESS classes are not included into the json file.
#   IGNORES classes are included with `ignore=True`.

# python tools/convert_datasets/mot2coco.py -i /apsara/yunji.cjy/tracking/MOT20 -o /apsara/yunji.cjy/tracking/MOT20/annotations --split-train --convert-det

import argparse
import os
import os.path as osp
from collections import defaultdict

import mmcv
import numpy as np
from tqdm import tqdm

USELESS = [3, 4, 5, 6, 9, 10, 11]
IGNORES = [2, 7, 8, 12, 13]


def parse_args():
    parser = argparse.ArgumentParser(
        description='Convert MOT label and detections to COCO-VID format.')
    parser.add_argument('-i', '--input', help='path of MOT data')
    parser.add_argument(
        '-o', '--output', help='path to save coco formatted label file')
    parser.add_argument(
        '--convert-det',
        action='store_true',
        help='convert official detection results.')
    parser.add_argument(
        '--split-train',
        action='store_true',
        help='split the train set into half-train and half-validate.')
    return parser.parse_args()


def parse_gts(gts, is_mot15):
    outputs = defaultdict(list)
    for gt in gts:
        gt = gt.strip().split(',')
        frame_id, ins_id = map(int, gt[:2])
        bbox = list(map(float, gt[2:6]))
        if is_mot15:
            conf = 1.
            class_id = 1
            visibility = 1.
        else:
            conf = float(gt[6])
            class_id = int(gt[7])
            visibility = float(gt[8])
        if class_id in USELESS:
            continue
        elif class_id in IGNORES:
            continue
        anns = dict(
            category_id=1,
            bbox=bbox,
            area=bbox[2] * bbox[3],
            iscrowd=False,
            visibility=visibility,
            mot_instance_id=ins_id,
            mot_conf=conf,
            mot_class_id=class_id)
        outputs[frame_id].append(anns)
    return outputs


def parse_dets(dets):
    outputs = defaultdict(list)
    for det in dets:
        det = det.strip().split(',')
        frame_id, ins_id = map(int, det[:2])
        assert ins_id == -1
        bbox = list(map(float, det[2:7]))
        # [x1, y1, x2, y2] to be consistent with mmdet
        bbox = [
            bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3], bbox[4]
        ]
        outputs[frame_id].append(bbox)

    return outputs


def main():
    args = parse_args()
    if not osp.isdir(args.output):
        os.makedirs(args.output)

    sets = ['train', 'test']
    if args.split_train:
        sets += ['half-train', 'half-val']
    vid_id, img_id, ann_id = 1, 1, 1

    for subset in sets:
        ins_id = 0
        print(f'Converting {subset} set to COCO format')
        if 'half' in subset:
            in_folder = osp.join(args.input, 'train')
        else:
            in_folder = osp.join(args.input, subset)
        out_file = osp.join(args.output, f'{subset}_cocoformat.json')
        outputs = defaultdict(list)
        outputs['categories'] = [dict(id=1, name='pedestrian')]
        if args.convert_det:
            det_file = osp.join(args.output, f'{subset}_detections.pkl')
            detections = dict(det_bboxes=dict())
        video_names = os.listdir(in_folder)
        for video_name in tqdm(video_names):
            # basic params
            parse_gt = 'test' not in subset
            ins_maps = dict()
            # load video infos
            video_folder = osp.join(in_folder, video_name)
            infos = mmcv.list_from_file(f'{video_folder}/seqinfo.ini')
            # video-level infos
            assert video_name == infos[1].strip().split('=')[1]
            img_folder = infos[2].strip().split('=')[1]
            img_names = os.listdir(f'{video_folder}/{img_folder}')
            img_names = sorted(img_names)
            fps = int(infos[3].strip().split('=')[1])
            num_imgs = int(infos[4].strip().split('=')[1])
            assert num_imgs == len(img_names)
            width = int(infos[5].strip().split('=')[1])
            height = int(infos[6].strip().split('=')[1])
            video = dict(
                id=vid_id,
                name=video_name,
                fps=fps,
                width=width,
                height=height)
            # parse annotations
            if parse_gt:
                gts = mmcv.list_from_file(f'{video_folder}/gt/gt.txt')
                if 'MOT15' in video_folder:
                    img2gts = parse_gts(gts, True)
                else:
                    img2gts = parse_gts(gts, False)
            if args.convert_det:
                dets = mmcv.list_from_file(f'{video_folder}/det/det.txt')
                img2dets = parse_dets(dets)
            # make half sets
            if 'half' in subset:
                split_frame = num_imgs // 2 + 1
                if 'train' in subset:
                    img_names = img_names[:split_frame]
                elif 'val' in subset:
                    img_names = img_names[split_frame:]
                else:
                    raise ValueError(
                        'subset must be named with `train` or `val`')
                mot_frame_ids = [str(int(_.split('.')[0])) for _ in img_names]
                with open(f'{video_folder}/gt/gt_{subset}.txt', 'wt') as f:
                    for gt in gts:
                        if gt.split(',')[0] in mot_frame_ids:
                            f.writelines(f'{gt}\n')
            # image and box level infos
            for frame_id, name in enumerate(img_names):
                img_name = osp.join(video_name, img_folder, name)
                mot_frame_id = int(name.split('.')[0])
                image = dict(
                    id=img_id,
                    video_id=vid_id,
                    file_name=img_name,
                    height=height,
                    width=width,
                    frame_id=frame_id,
                    mot_frame_id=mot_frame_id)
                if parse_gt:
                    gts = img2gts[mot_frame_id]
                    for gt in gts:
                        gt.update(id=ann_id, image_id=img_id)
                        mot_ins_id = gt['mot_instance_id']
                        if mot_ins_id in ins_maps:
                            gt['instance_id'] = ins_maps[mot_ins_id]
                        else:
                            gt['instance_id'] = ins_id
                            ins_maps[mot_ins_id] = ins_id
                            ins_id += 1
                        outputs['annotations'].append(gt)
                        ann_id += 1
                if args.convert_det:
                    dets = np.array(img2dets[mot_frame_id])
                    if dets.ndim == 1:
                        assert len(dets) == 0
                        dets = np.zeros((0, 5))
                    detections['det_bboxes'][img_name] = [dets]
                outputs['images'].append(image)
                img_id += 1
            outputs['videos'].append(video)
            vid_id += 1
            outputs['num_instances'] = ins_id
        print(f'{subset} has {ins_id} instances.')
        mmcv.dump(outputs, out_file)
        if args.convert_det:
            mmcv.dump(detections, det_file)
            print(f'Done! Saved as {out_file} and {det_file}')
        else:
            print(f'Done! Saved as {out_file}')


if __name__ == '__main__':
    main()



# Copyright (c) Alibaba, Inc. and its affiliates.
import torch

from .base_evaluator import Evaluator
from .builder import EVALUATORS
from .metric_registry import METRICS


@EVALUATORS.register_module
class MSEEvaluator(Evaluator):
    """ MSEEvaluator evaluator,

  """

    def __init__(self,
                 dataset_name=None,
                 metric_names=['avg_mse'],
                 neck_num=None):
        '''
    '''

        self.metric = 'min'
        self.dataset_name = dataset_name
        self.neck_num = neck_num

        super(MSEEvaluator, self).__init__(dataset_name, metric_names)

    def _evaluate_impl(self, results, gt_label):
        """ Retrival evaluate do the topK retrival, by measuring the distance of every 1 vs other.
        get the topK nearest, and count the match of ID. if Retrival = 1, Miss = 0. Finally average all
        RetrivalRate.
        """
        # first print() is to show shape clearly in multi-process situation. don't comment it
        print()

        if self.neck_num is None:
            try:
                results = results.cuda()
            except:
                results = results['neck'].cuda()
        else:
            results = results['neck_%d_0' % self.neck_num].cuda()

        gt_label = gt_label.cuda()

        # print(results.shape)
        # print(gt_label.shape)

        if results.shape[1] > 1:
            n, c = results.size()
            prob = torch.nn.Softmax(dim=1)(results)

            distribute = torch.arange(0, c).repeat(n, 1).to(results.device)
            results = (distribute * prob).sum(dim=1)

        eval_res = {}
        avg_mse = torch.mean(torch.abs(results - gt_label))
        eval_res['avg_mse'] = avg_mse.item()

        return eval_res


METRICS.register_default_best_metric(MSEEvaluator, 'avg_mse', 'max')



import torch.nn as nn

from ..registry import BACKBONES


@BACKBONES.register_module
class BenchMarkMLP(nn.Module):

    def __init__(self,
                 feature_num,
                 num_classes=1000,
                 avg_pool=False,
                 **kwargs):
        super(BenchMarkMLP, self).__init__()

        self.fc1 = nn.Linear(feature_num, feature_num)
        self.relu1 = nn.ReLU()
        self.dropout = nn.Dropout(p=0.5)
        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        self.avg_pool = avg_pool

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(
                    m.weight, mode='fan_out', nonlinearity='relu')

    def forward(self, x):
        if self.avg_pool:
            x = self.pool(x)
        x = self.fc1(x)
        x = self.relu1(x)
        return tuple([x])



# Copyright (c) Alibaba, Inc. and its affiliates.
from .builder import build_match_cost
from .match_cost import BBox3DL1Cost, FocalLossCost, IoUCost

__all__ = ['BBox3DL1Cost', 'FocalLossCost', 'IoUCost', 'build_match_cost']



import argparse
import os
from shutil import copyfile


def generate_query(download_path, save_path):
    # -----------------------------------------
    # query
    query_path = os.path.join(download_path, 'query')
    query_save_path = os.path.join(save_path, 'query')
    if not os.path.isdir(query_save_path):
        os.mkdir(query_save_path)

    for root, dirs, files in os.walk(query_path, topdown=True):
        for name in files:
            if not name[-3:] == 'jpg':
                continue
            ID = name.split('_')
            src_path = os.path.join(query_path, name)
            dst_path = os.path.join(query_save_path, ID[0])
            if not os.path.isdir(dst_path):
                os.mkdir(dst_path)
            copyfile(src_path, os.path.join(dst_path, name))


def generate_multi_query(download_path, save_path):
    # -----------------------------------------
    # multi-query
    query_path = os.path.join(download_path, 'gt_bbox')
    # for dukemtmc-reid, we do not need multi-query
    if os.path.isdir(query_path):
        query_save_path = os.path.join(save_path, 'multi-query')
        if not os.path.isdir(query_save_path):
            os.mkdir(query_save_path)

        for root, dirs, files in os.walk(query_path, topdown=True):
            for name in files:
                if not name[-3:] == 'jpg':
                    continue
                ID = name.split('_')
                src_path = os.path.join(query_path, name)
                dst_path = os.path.join(query_save_path, ID[0])
                if not os.path.isdir(dst_path):
                    os.mkdir(dst_path)
                copyfile(src_path, os.path.join(dst_path, name))


def generate_gallery(download_path, save_path):
    # -----------------------------------------
    # gallery
    gallery_path = os.path.join(download_path, 'bounding_box_test')
    gallery_save_path = os.path.join(save_path, 'gallery')
    if not os.path.isdir(gallery_save_path):
        os.mkdir(gallery_save_path)

    for root, dirs, files in os.walk(gallery_path, topdown=True):
        for name in files:
            if not name[-3:] == 'jpg':
                continue
            ID = name.split('_')
            src_path = os.path.join(gallery_path, name)
            dst_path = os.path.join(gallery_save_path, ID[0])
            if not os.path.isdir(dst_path):
                os.mkdir(dst_path)
            copyfile(src_path, os.path.join(dst_path, name))


def generate_train_val(download_path, save_path):
    label_dict = dict()
    # ---------------------------------------
    # train_all
    train_path = os.path.join(download_path, 'bounding_box_train')
    train_save_path = os.path.join(save_path, 'train_all')
    txt_base = os.path.join(save_path, 'meta')
    if not os.path.isdir(train_save_path):
        os.mkdir(train_save_path)
    if not os.path.isdir(txt_base):
        os.mkdir(txt_base)
    train_txt_path = open(os.path.join(txt_base, 'train_all.txt'), 'w')
    label_map_path = open(os.path.join(txt_base, 'label_map.txt'), 'w')

    for root, dirs, files in os.walk(train_path, topdown=True):
        for name in files:
            if not name[-3:] == 'jpg':
                continue
            ID = name.split('_')
            src_path = os.path.join(train_path, name)
            dst_path = os.path.join(train_save_path, ID[0])
            if ID[0] not in label_dict:
                label_dict[ID[0]] = str(len(label_dict))
                label_map_path.write(ID[0] + ' ' + label_dict[ID[0]] + '\n')
            train_txt_path.write(dst_path + '/' + name + ' ' +
                                 label_dict[ID[0]] + '\n')
            if not os.path.isdir(dst_path):
                os.mkdir(dst_path)
            copyfile(src_path, os.path.join(dst_path, name))

    # ---------------------------------------
    # train_val
    train_path = os.path.join(download_path, 'bounding_box_train')
    train_save_path = os.path.join(save_path, 'train')
    val_save_path = os.path.join(save_path, 'val')
    train_txt_path = open(os.path.join(txt_base, 'train.txt'), 'w')
    val_txt_path = open(os.path.join(txt_base, 'val.txt'), 'w')
    if not os.path.isdir(train_save_path):
        os.mkdir(train_save_path)
        os.mkdir(val_save_path)

    for root, dirs, files in os.walk(train_path, topdown=True):
        for name in files:
            if not name[-3:] == 'jpg':
                continue
            ID = name.split('_')
            src_path = os.path.join(train_path, name)
            train_dst_path = os.path.join(train_save_path, ID[0])
            val_dst_path = os.path.join(val_save_path, ID[0])
            if ID[0] not in label_dict:
                label_dict[ID[0]] = str(len(label_dict))
                label_map_path.write(ID[0] + ' ' + label_dict[ID[0]] + '\n')
            if not os.path.isdir(
                    val_dst_path):  # first image is used as val image
                os.mkdir(val_dst_path)
                copyfile(src_path, os.path.join(val_dst_path, name))
                val_txt_path.write(val_dst_path + '/' + name + ' ' +
                                   label_dict[ID[0]] + '\n')
            else:
                if not os.path.isdir(train_dst_path):
                    os.mkdir(train_dst_path)
                copyfile(src_path, os.path.join(train_dst_path, name))
                train_txt_path.write(train_dst_path + '/' + name + ' ' +
                                     label_dict[ID[0]] + '\n')


def main(download_path):
    if not os.path.isdir(download_path):
        print('please change the download_path')

    save_path = download_path + '/pytorch'
    if not os.path.isdir(save_path):
        os.mkdir(save_path)

    generate_gallery(download_path, save_path)
    generate_multi_query(download_path, save_path)
    generate_query(download_path, save_path)
    generate_train_val(download_path, save_path)


def parse_args():
    parser = argparse.ArgumentParser(description='prepare market1501 datasets')
    parser.add_argument(
        '--download-path',
        type=str,
        default='./data/Market1501',
        help='You only need to change this line to your dataset download path')
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()
    main(args.download_path)



# Copyright (c) Alibaba, Inc. and its affiliates.
from distutils.version import LooseVersion

import cv2
import numpy as np
import torch
from PIL import Image
from timm.data import create_transform
from torchvision import transforms as _transforms

from easycv.datasets.registry import PIPELINES
from easycv.utils.registry import build_from_cfg


@PIPELINES.register_module
class MAEFtAugment(object):
    """RandAugment data augmentation method based on
    `"RandAugment: Practical automated data augmentation
    with a reduced search space"
    <https://arxiv.org/abs/1909.13719>`_.
    This code is borrowed from <https://github.com/pengzhiliang/MAE-pytorch>
    Args:
        input_size(int): images input size
        color_jitter(float): Color jitter factor
        auto_augment: Use AutoAugment policy
        iterpolation: Training interpolation
        re_prob: Random erase prob
        re_mode: Random erase mode
        re_count: Random erase count
        mean: mean used for normalization
        std: std used for normalization
        is_train: If True use all augmentation strategy
    """

    def __init__(self,
                 input_size=None,
                 color_jitter=None,
                 auto_augment=None,
                 interpolation=None,
                 re_prob=None,
                 re_mode=None,
                 re_count=None,
                 mean=None,
                 std=None,
                 is_train=True):
        resize_im = input_size > 32
        if is_train:
            self.trans = create_transform(
                input_size=input_size,
                is_training=True,
                color_jitter=color_jitter,
                auto_augment=auto_augment,
                interpolation=interpolation,
                re_prob=re_prob,
                re_mode=re_mode,
                re_count=re_count,
                mean=mean,
                std=std,
            )
        else:
            t = []
            if resize_im:
                if input_size < 384:
                    crop_pct = 224 / 256
                else:
                    crop_pct = 1.0
                size = int(input_size / crop_pct)
                t.append(
                    _transforms.Resize(
                        size, interpolation=3
                    ),  # to maintain same ratio w.r.t. 224 images
                )
                t.append(_transforms.CenterCrop(input_size))
            t.append(_transforms.ToTensor())
            t.append(_transforms.Normalize(mean, std))
            self.trans = _transforms.Compose(t)

    def __call__(self, results):
        img = results['img']
        img = self.trans(img)
        results['img'] = img
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        return repr_str


class _RandomApply(_transforms.RandomApply):

    def forward(self, results):
        if self.p < torch.rand(1):
            return results
        for t in self.transforms:
            results = t(results)
        return results


@PIPELINES.register_module
class RandomAppliedTrans(object):
    '''Randomly applied transformations.
    Args:
        transforms (List[Dict]): List of transformations in dictionaries.
    '''

    def __init__(self, transforms, p=0.5):
        t = [build_from_cfg(t, PIPELINES) for t in transforms]
        self.trans = _RandomApply(t, p=p)

    def __call__(self, results):
        return self.trans(results)

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str


# custom transforms
@PIPELINES.register_module
class Lighting(object):
    """Lighting noise(AlexNet - style PCA - based noise)"""
    _IMAGENET_PCA = {
        'eigval':
        torch.Tensor([0.2175, 0.0188, 0.0045]),
        'eigvec':
        torch.Tensor([
            [-0.5675, 0.7192, 0.4009],
            [-0.5808, -0.0045, -0.8140],
            [-0.5836, -0.6948, 0.4203],
        ])
    }

    def __init__(self):
        self.alphastd = 0.1
        self.eigval = self._IMAGENET_PCA['eigval']
        self.eigvec = self._IMAGENET_PCA['eigvec']

    def __call__(self, results):
        for key in results.get('img_fields', ['img']):
            img = results[key]
            assert isinstance(img, torch.Tensor), \
                'Expect torch.Tensor, got {}'.format(type(img))
            if self.alphastd == 0:
                continue

            alpha = img.new().resize_(3).normal_(0, self.alphastd)
            rgb = self.eigvec.type_as(img).clone()\
                .mul(alpha.view(1, 3).expand(3, 3))\
                .mul(self.eigval.view(1, 3).expand(3, 3))\
                .sum(1).squeeze()

            results[key] = img.add(rgb.view(3, 1, 1).expand_as(img))

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str


if LooseVersion(torch.__version__) < LooseVersion('1.7.0'):

    @PIPELINES.register_module
    class GaussianBlur(object):

        def __init__(self, kernel_size, sigma=(0.1, 2.0)):
            self.sigma_min = sigma[0]
            self.sigma_max = sigma[1]
            self.kernel_size = kernel_size

        def __call__(self, results):
            for key in results.get('img_fields', ['img']):
                img = results[key]
                sigma = np.random.uniform(self.sigma_min, self.sigma_max)
                img = cv2.GaussianBlur(
                    np.array(img), (self.kernel_size, self.kernel_size), sigma)
                results[key] = Image.fromarray(img.astype(np.uint8))

            return results

        def __repr__(self):
            repr_str = self.__class__.__name__
            return repr_str


@PIPELINES.register_module
class Solarization(object):

    def __init__(self, threshold=128):
        self.threshold = threshold

    def __call__(self, results):
        for key in results.get('img_fields', ['img']):
            img = results[key]
            img = np.array(img)
            img = np.where(img < self.threshold, img, 255 - img)
            results[key] = Image.fromarray(img.astype(np.uint8))

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str



import copy
import math

import numpy as np
import torch
import torch.nn as nn


def conv_bn(inp, oup, kernel, stride, padding=1):
    return nn.Sequential(
        nn.Conv2d(inp, oup, kernel, stride, padding, bias=False),
        nn.BatchNorm2d(oup), nn.PReLU(oup))


def conv_no_relu(inp, oup, kernel, stride, padding=1):
    return nn.Sequential(
        nn.Conv2d(inp, oup, kernel, stride, padding, bias=False),
        nn.BatchNorm2d(oup))


class View(nn.Module):

    def __init__(self, shape):
        super(View, self).__init__()
        self.shape = shape

    def forward(self, x):
        return x.view(*self.shape)


class Softmax(nn.Module):

    def __init__(self, dim):
        super(Softmax, self).__init__()
        self.softmax = nn.Softmax(dim)

    def forward(self, x):
        return self.softmax(x)


class InvertedResidual(nn.Module):

    def __init__(self,
                 inp,
                 oup,
                 kernel_size,
                 stride,
                 padding,
                 expand_ratio=2,
                 use_connect=False,
                 activation='relu'):
        super(InvertedResidual, self).__init__()

        hid_channels = int(inp * expand_ratio)
        if activation == 'relu':
            self.conv = nn.Sequential(
                nn.Conv2d(inp, hid_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hid_channels), nn.ReLU(inplace=True),
                nn.Conv2d(
                    hid_channels,
                    hid_channels,
                    kernel_size,
                    stride,
                    padding,
                    groups=hid_channels,
                    bias=False), nn.BatchNorm2d(hid_channels),
                nn.ReLU(inplace=True),
                nn.Conv2d(hid_channels, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup))
        elif activation == 'prelu':
            self.conv = nn.Sequential(
                nn.Conv2d(inp, hid_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hid_channels), nn.PReLU(hid_channels),
                nn.Conv2d(
                    hid_channels,
                    hid_channels,
                    kernel_size,
                    stride,
                    padding,
                    groups=hid_channels,
                    bias=False), nn.BatchNorm2d(hid_channels),
                nn.PReLU(hid_channels),
                nn.Conv2d(hid_channels, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup))
        elif activation == 'half_v1':
            self.conv = nn.Sequential(
                nn.Conv2d(inp, hid_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hid_channels), nn.ReLU(inplace=True),
                nn.Conv2d(
                    hid_channels,
                    hid_channels,
                    kernel_size,
                    stride,
                    padding,
                    groups=hid_channels,
                    bias=False), nn.BatchNorm2d(hid_channels),
                nn.PReLU(hid_channels),
                nn.Conv2d(hid_channels, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup))
        elif activation == 'half_v2':
            self.conv = nn.Sequential(
                nn.Conv2d(inp, hid_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hid_channels), nn.PReLU(hid_channels),
                nn.Conv2d(
                    hid_channels,
                    hid_channels,
                    kernel_size,
                    stride,
                    padding,
                    groups=hid_channels,
                    bias=False), nn.BatchNorm2d(hid_channels),
                nn.ReLU(inplace=True),
                nn.Conv2d(hid_channels, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup))
        self.use_connect = use_connect

    def forward(self, x):
        if self.use_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class Residual(nn.Module):

    def __init__(self,
                 inp,
                 oup,
                 kernel_size,
                 stride,
                 padding,
                 use_connect=False,
                 activation='relu'):
        super(Residual, self).__init__()

        self.use_connect = use_connect

        if activation == 'relu':
            self.conv = nn.Sequential(
                nn.Conv2d(
                    inp,
                    inp,
                    kernel_size,
                    stride,
                    padding,
                    groups=inp,
                    bias=False), nn.BatchNorm2d(inp), nn.ReLU(inplace=True),
                nn.Conv2d(inp, oup, 1, 1, 0, bias=False), nn.BatchNorm2d(oup),
                nn.ReLU(inplace=True))
        elif activation == 'prelu':
            self.conv = nn.Sequential(
                nn.Conv2d(
                    inp,
                    inp,
                    kernel_size,
                    stride,
                    padding,
                    groups=inp,
                    bias=False), nn.BatchNorm2d(inp), nn.PReLU(inp),
                nn.Conv2d(inp, oup, 1, 1, 0, bias=False), nn.BatchNorm2d(oup),
                nn.PReLU(oup))
        elif activation == 'half_v1':
            self.conv = nn.Sequential(
                nn.Conv2d(
                    inp,
                    inp,
                    kernel_size,
                    stride,
                    padding,
                    groups=inp,
                    bias=False), nn.BatchNorm2d(inp), nn.ReLU(inplace=True),
                nn.Conv2d(inp, oup, 1, 1, 0, bias=False), nn.BatchNorm2d(oup),
                nn.PReLU(oup))
        elif activation == 'half_v2':
            self.conv = nn.Sequential(
                nn.Conv2d(
                    inp,
                    inp,
                    kernel_size,
                    stride,
                    padding,
                    groups=inp,
                    bias=False), nn.BatchNorm2d(inp), nn.PReLU(inp),
                nn.Conv2d(inp, oup, 1, 1, 0, bias=False), nn.BatchNorm2d(oup),
                nn.ReLU(inplace=True))

    def forward(self, x):
        if self.use_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


def pose_accuracy(output, target):
    with torch.no_grad():
        output = output.detach().cpu().numpy()
        target = target.detach().cpu().numpy()

        acc = np.mean(np.abs(output - target))
        return acc


def ION(output, target, left_eye_left_coner_idx, right_eye_right_corner_idx,
        num_pts):
    with torch.no_grad():
        output = output.view(-1, num_pts, 2).cpu().numpy()
        target = target.view(-1, num_pts, 2).cpu().numpy()

        interocular = target[:,
                             left_eye_left_coner_idx] - target[:,
                                                               right_eye_right_corner_idx]
        interocular = np.sqrt(
            np.square(interocular[:, 0]) + np.square(interocular[:, 1])) + 1e-5
        dist = target - output
        dist = np.sqrt(np.square(dist[:, :, 0]) + np.square(dist[:, :, 1]))
        dist = np.sum(dist, axis=1)
        nme = dist / (interocular * num_pts)

    return np.mean(nme)


def get_keypoint_accuracy(output, target_point):
    accuracy = dict()
    num_points = 106
    left_eye_left_corner_index = 66
    right_eye_right_corner_index = 79

    nme = ION(output, target_point, left_eye_left_corner_index,
              right_eye_right_corner_index, num_points)

    accuracy['nme'] = nme

    return accuracy


def get_pose_accuracy(output, target_pose):
    accuracy = dict()
    pose_acc = pose_accuracy(output, target_pose)
    accuracy['pose_acc'] = float(pose_acc)
    return accuracy



# Copyright (c) Alibaba, Inc. and its affiliates.
from modelscope.models.builder import MODELS
from modelscope.utils.constant import Tasks

from easycv.models.face.face_keypoint import FaceKeypoint
from easycv.toolkit.modelscope.metainfo import EasyCVModels as Models
from easycv.toolkit.modelscope.models.base import EasyCVBaseModel


@MODELS.register_module(
    group_key=Tasks.face_2d_keypoints, module_name=Models.face_2d_keypoints)
class Face2DKeypoints(EasyCVBaseModel, FaceKeypoint):

    def __init__(self, model_dir=None, *args, **kwargs):
        EasyCVBaseModel.__init__(self, model_dir, args, kwargs)
        FaceKeypoint.__init__(self, *args, **kwargs)



# Copyright (c) 2022 IDEA. All Rights Reserved.
# Copyright (c) Alibaba, Inc. and its affiliates.
import math

import torch
import torch.nn as nn

from easycv.models.builder import HEADS, build_neck
from easycv.models.detection.utils import (DetrPostProcess, box_xyxy_to_cxcywh,
                                           inverse_sigmoid)
from easycv.models.loss import DNCriterion, HungarianMatcher, SetCriterion
from easycv.models.utils import MLP
from .dn_components import dn_post_process, prepare_for_dn


@HEADS.register_module()
class DABDETRHead(nn.Module):
    """Implements the DAB-DETR head.
    See `paper: DAB-DETR: Dynamic Anchor Boxes are Better Queries for DETR
    <https://arxiv.org/abs/2201.12329> and DN-DETR: Accelerate DETR Training by Introducing Query DeNoising
    <https://arxiv.org/abs/2203.01305>`_ for details.
    Args:
        num_classes (int): Number of categories excluding the background.
    """

    def __init__(self,
                 num_classes,
                 embed_dims,
                 query_dim=4,
                 iter_update=True,
                 num_queries=300,
                 num_select=300,
                 random_refpoints_xy=False,
                 num_patterns=0,
                 bbox_embed_diff_each_layer=False,
                 dn_components=None,
                 transformer=None,
                 cost_dict={
                     'cost_class': 1,
                     'cost_bbox': 5,
                     'cost_giou': 2,
                 },
                 weight_dict={
                     'loss_ce': 1,
                     'loss_bbox': 5,
                     'loss_giou': 2
                 },
                 **kwargs):

        super(DABDETRHead, self).__init__()

        self.matcher = HungarianMatcher(
            cost_dict=cost_dict, cost_class_type='focal_loss_cost')
        self.criterion = SetCriterion(
            num_classes,
            matcher=self.matcher,
            weight_dict=weight_dict,
            losses=['labels', 'boxes'],
            loss_class_type='focal_loss')
        if dn_components is not None:
            self.dn_criterion = DNCriterion(weight_dict)
        self.postprocess = DetrPostProcess(num_select=num_select)
        self.transformer = build_neck(transformer)

        self.class_embed = nn.Linear(embed_dims, num_classes)
        if bbox_embed_diff_each_layer:
            self.bbox_embed = nn.ModuleList(
                [MLP(embed_dims, embed_dims, 4, 3) for i in range(6)])
        else:
            self.bbox_embed = MLP(embed_dims, embed_dims, 4, 3)
        if iter_update:
            self.transformer.decoder.bbox_embed = self.bbox_embed

        self.num_classes = num_classes
        self.num_queries = num_queries
        self.embed_dims = embed_dims
        self.query_dim = query_dim
        self.bbox_embed_diff_each_layer = bbox_embed_diff_each_layer
        self.dn_components = dn_components

        self.query_embed = nn.Embedding(num_queries, query_dim)
        self.random_refpoints_xy = random_refpoints_xy
        if random_refpoints_xy:
            self.query_embed.weight.data[:, :2].uniform_(0, 1)
            self.query_embed.weight.data[:, :2] = inverse_sigmoid(
                self.query_embed.weight.data[:, :2])
            self.query_embed.weight.data[:, :2].requires_grad = False

        self.num_patterns = num_patterns
        if not isinstance(num_patterns, int):
            Warning('num_patterns should be int but {}'.format(
                type(num_patterns)))
            self.num_patterns = 0
        if self.num_patterns > 0:
            self.patterns = nn.Embedding(self.num_patterns, embed_dims)

        if self.dn_components:
            # leave one dim for indicator
            self.label_enc = nn.Embedding(num_classes + 1, embed_dims - 1)

    def init_weights(self):
        self.transformer.init_weights()

        # init prior_prob setting for focal loss
        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        self.class_embed.bias.data = torch.ones(self.num_classes) * bias_value

        # import ipdb; ipdb.set_trace()
        # init bbox_embed
        if self.bbox_embed_diff_each_layer:
            for bbox_embed in self.bbox_embed:
                nn.init.constant_(bbox_embed.layers[-1].weight.data, 0)
                nn.init.constant_(bbox_embed.layers[-1].bias.data, 0)
        else:
            nn.init.constant_(self.bbox_embed.layers[-1].weight.data, 0)
            nn.init.constant_(self.bbox_embed.layers[-1].bias.data, 0)

    def prepare(self, feats, targets=None, mode='train'):
        bs = feats[0].shape[0]
        query_embed = self.query_embed.weight
        if self.dn_components:
            # default pipeline
            self.dn_components['num_patterns'] = self.num_patterns
            self.dn_components['targets'] = targets
            # prepare for dn
            tgt, query_embed, attn_mask, mask_dict = prepare_for_dn(
                mode, self.dn_components, query_embed, bs, self.num_queries,
                self.num_classes, self.embed_dims, self.label_enc)
            if self.num_patterns > 0:
                l = tgt.shape[0]
                tgt[l - self.num_queries * self.num_patterns:] += \
                    self.patterns.weight[:, None, None, :].repeat(1, self.num_queries, bs, 1).flatten(0, 1)
            return query_embed, tgt, attn_mask, mask_dict
        else:
            query_embed = query_embed.unsqueeze(1).repeat(1, bs, 1)
            if self.num_patterns == 0:
                tgt = torch.zeros(
                    self.num_queries,
                    bs,
                    self.embed_dims,
                    device=query_embed.device)
            else:
                tgt = self.patterns.weight[:, None, None, :].repeat(
                    1, self.num_queries, bs,
                    1).flatten(0, 1)  # n_q*n_pat, bs, d_model
                query_embed = query_embed.repeat(self.num_patterns, 1,
                                                 1)  # n_q*n_pat, bs, d_model
        return query_embed, tgt, None, None

    def forward(self,
                feats,
                img_metas,
                query_embed=None,
                tgt=None,
                attn_mask=None,
                mask_dict=None):
        """Forward function.
        Args:
            feats (tuple[Tensor]): Features from the upstream network, each is
                a 4D-tensor.
            img_metas (list[dict]): List of image information.
        Returns:
            tuple[list[Tensor], list[Tensor]]: Outputs for all scale levels.
                - all_cls_scores_list (list[Tensor]): Classification scores \
                    for each scale level. Each is a 4D-tensor with shape \
                    [nb_dec, bs, num_query, cls_out_channels]. Note \
                    `cls_out_channels` should includes background.
                - all_bbox_preds_list (list[Tensor]): Sigmoid regression \
                    outputs for each scale level. Each is a 4D-tensor with \
                    normalized coordinate format (cx, cy, w, h) and shape \
                    [nb_dec, bs, num_query, 4].
        """

        feats = self.transformer(
            feats, img_metas, query_embed, tgt, attn_mask=attn_mask)

        hs, reference = feats
        outputs_class = self.class_embed(hs)
        if not self.bbox_embed_diff_each_layer:
            reference_before_sigmoid = inverse_sigmoid(reference)
            tmp = self.bbox_embed(hs)
            tmp[..., :self.query_dim] += reference_before_sigmoid
            outputs_coord = tmp.sigmoid()
        else:
            reference_before_sigmoid = inverse_sigmoid(reference)
            outputs_coords = []
            for lvl in range(hs.shape[0]):
                tmp = self.bbox_embed[lvl](hs[lvl])
                tmp[..., :self.query_dim] += reference_before_sigmoid[lvl]
                outputs_coord = tmp.sigmoid()
                outputs_coords.append(outputs_coord)
            outputs_coord = torch.stack(outputs_coords)

        if mask_dict is not None:
            # dn post process
            outputs_class, outputs_coord = dn_post_process(
                outputs_class, outputs_coord, mask_dict)
        out = {
            'pred_logits': outputs_class[-1],
            'pred_boxes': outputs_coord[-1]
        }

        out['aux_outputs'] = self._set_aux_loss(outputs_class, outputs_coord)
        return out

    @torch.jit.unused
    def _set_aux_loss(self, outputs_class, outputs_coord):
        # this is a workaround to make torchscript happy, as torchscript
        # doesn't support dictionary with non-homogeneous values, such
        # as a dict having both a Tensor and a list.
        return [{
            'pred_logits': a,
            'pred_boxes': b
        } for a, b in zip(outputs_class[:-1], outputs_coord[:-1])]

    # over-write because img_metas are needed as inputs for bbox_head.
    def forward_train(self, x, img_metas, gt_bboxes, gt_labels):
        """Forward function for training mode.
        Args:
            x (list[Tensor]): Features from backbone.
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes (Tensor): Ground truth bboxes of the image,
                shape (num_gts, 4).
            gt_labels (Tensor): Ground truth labels of each box,
                shape (num_gts,).
            gt_bboxes_ignore (Tensor): Ground truth bboxes to be
                ignored, shape (num_ignored_gts, 4).
            proposal_cfg (mmcv.Config): Test / postprocessing configuration,
                if None, test_cfg would be used.
        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        # prepare ground truth
        for i in range(len(img_metas)):
            img_h, img_w, _ = img_metas[i]['img_shape']
            # DETR regress the relative position of boxes (cxcywh) in the image.
            # Thus the learning target should be normalized by the image size, also
            # the box format should be converted from defaultly x1y1x2y2 to cxcywh.
            factor = gt_bboxes[i].new_tensor([img_w, img_h, img_w,
                                              img_h]).unsqueeze(0)
            gt_bboxes[i] = box_xyxy_to_cxcywh(gt_bboxes[i]) / factor

        targets = []
        for gt_label, gt_bbox in zip(gt_labels, gt_bboxes):
            targets.append({'labels': gt_label, 'boxes': gt_bbox})

        query_embed, tgt, attn_mask, mask_dict = self.prepare(
            x, targets=targets, mode='train')

        outputs = self.forward(
            x,
            img_metas,
            query_embed=query_embed,
            tgt=tgt,
            attn_mask=attn_mask,
            mask_dict=mask_dict)

        losses = self.criterion(outputs, targets)
        if self.dn_components:
            losses.update(
                self.dn_criterion(mask_dict, len(outputs['aux_outputs'])))

        return losses

    def forward_test(self, x, img_metas):
        query_embed, tgt, attn_mask, mask_dict = self.prepare(x, mode='test')

        outputs = self.forward(
            x,
            img_metas,
            query_embed=query_embed,
            tgt=tgt,
            attn_mask=attn_mask,
            mask_dict=mask_dict)

        ori_shape_list = []
        for i in range(len(img_metas)):
            ori_h, ori_w, _ = img_metas[i]['ori_shape']
            ori_shape_list.append(torch.as_tensor([ori_h, ori_w]))
        orig_target_sizes = torch.stack(ori_shape_list, dim=0)

        results = self.postprocess(outputs, orig_target_sizes, img_metas)
        return results



# Copyright (c) Alibaba, Inc. and its affiliates.
from .det_transform import (EastRandomCropData, IaaAugment, MakeBorderMap,
                            MakeShrinkMap, OCRDetResize)
from .label_ops import CTCLabelEncode, MultiLabelEncode, SARLabelEncode
from .rec_transform import ClsResizeImg, RecAug, RecConAug, RecResizeImg



_base_ = [
    './deitiii_base_patch16_192.py',
    './threeaug_imagenet_classification_224.py', './deitiii_schedule.py'
]
# model settings
model = dict(
    backbone=dict(
        type='VisionTransformer',
        img_size=[224],
        embed_dim=384,
        num_heads=6,
        drop_path_rate=0.05))

# optimizer
optimizer = dict(lr=0.004)

lr_config = dict(
    min_lr_ratio=0.00001 / 0.004,
    warmup_ratio=0.000001 / 0.004,
)



#!/usr/bin/env python3
# Copyright (c) 2014-2021 Megvii Inc. All rights reserved.
from .boxes import (batched_nms, bbox2result, bbox_overlaps, bboxes_iou,
                    box_cxcywh_to_xyxy, box_iou, box_xyxy_to_cxcywh,
                    distance2bbox, fp16_clamp, generalized_box_iou)
from .generator import MlvlPointGenerator
from .misc import (accuracy, gen_encoder_output_proposals,
                   gen_sineembed_for_position, interpolate, inverse_sigmoid,
                   select_single_mlvl)
from .postprocess import DetrPostProcess, output_postprocess, postprocess



_base_ = 'configs/base.py'

# Studio2.0 config

# dataaug
img_size = 224

# channel setup
backbone_channel = 384
feature_channel = 1536
num_classes = 1000

# supported_backbone

# supported_softmax+loss ,all this loss use same parameters
# [
#    "ModelParallelSoftmaxLoss",    m=28.6, s=64
#    "ModelParallelAMSoftmaxLoss" m=0.4, s=30
# ]

metric_loss_name = 'ModelParallelSoftmaxLoss'
metric_loss_weight = 1.0
metric_loss_scale = 30
metric_loss_margin = 0.4

evaluator1_name = 'ClsEvaluator'
evaluator2_name = 'RetrivalTopKEvaluator'

# ----------------------------- original config ------------------------------------#
log_config = dict(
    interval=1, hooks=[
        dict(type='TextLoggerHook'),
    ])

# oss io config
oss_io_config = dict(
    ak_id='your oss ak id',
    ak_secret='your oss ak secret',
    hosts='oss-cn-zhangjiakou.aliyuncs.com',
    buckets=['your_bucket'])

work_dir = 'oss://path/to/work_dirs/classification/'

# model settings
model = dict(
    type='Classification',
    backbone=dict(
        type='PytorchImageModelWrapper',

        # swin(dynamic)
        model_name='dynamic_swin_tiny_p4_w7_224',
        # model_name = 'dynamic_swin_small_p4_w7_224',
        # model_name = 'dynamic_swin_base_p4_w7_224',

        # deit(224)
        # model_name='vit_deit_tiny_distilled_patch16_224',   # good 192,
        # model_name='vit_deit_small_distilled_patch16_224',   # good 192,

        # xcit(224)
        # model_name='xcit_small_12_p16',
        # model_name='xcit_medium_24_p16',
        # model_name='xcit_large_24_p8'

        # resnet
        # model_name = 'resnet50',
        # model_name = 'resnet18',
        # model_name = 'resnet34',
        # model_name = 'resnet101',
        num_classes='${backbone_channel}',
    ),
    neck=dict(
        type='RetrivalNeck',
        in_channels='${backbone_channel}',
        out_channels='${feature_channel}',
        with_avg_pool=True,
        cdg_config=['G', 'S']),
    head=[
        # dict(
        #     type='MpMetrixHead',
        #     with_avg_pool=True,
        #     in_channels='${feature_channel}',
        #     loss_config = [
        #         dict(
        #             type="CrossEntropyLossWithLabelSmooth",
        #             loss_weight=1.0,
        #             norm=True,
        #             ddp=False,
        #             label_smooth=0.1,
        #             temperature=0.05,
        #             with_cls=True,
        #             embedding_size='${feature_channel}',
        #             num_classes='${num_classes}'
        #         )
        #     ],
        #     input_feature_index = [1]
        # ),
        dict(
            type='MpMetrixHead',
            with_avg_pool=True,
            in_channels='${feature_channel}',
            loss_config=[
                dict(
                    type='${metric_loss_name}',
                    loss_weight='${metric_loss_weight}',
                    norm=False,
                    ddp=True,
                    scale='${metric_loss_scale}',
                    margin='${metric_loss_margin}',
                    embedding_size='${feature_channel}',
                    num_classes='${num_classes}')
            ],
            input_feature_index=[0])
    ])

img_norm_cfg = dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
mean = [x * 255 for x in img_norm_cfg['mean']]
std = [x * 255 for x in img_norm_cfg['std']]
train_pipeline = [
    dict(type='DaliImageDecoder'),
    dict(
        type='DaliRandomResizedCrop',
        size='${img_size}',
        random_area=(0.2, 1.0)),
    dict(
        type='DaliCropMirrorNormalize',
        crop=['${img_size}', '${img_size}'],
        mean=mean,
        std=std,
        crop_pos_x=[0.0, 1.0],
        crop_pos_y=[0.0, 1.0],
        prob=0.5)
]
val_pipeline = [
    dict(type='DaliImageDecoder'),
    dict(type='DaliResize', resize_shorter='${img_size} * 1.15'),
    dict(
        type='DaliCropMirrorNormalize',
        crop=['${img_size}', '${img_size}'],
        mean=mean,
        std=std,
        prob=0.0)
]
data = dict(
    imgs_per_gpu=32,  # total 256
    workers_per_gpu=1,
    train=dict(
        type='DaliImageNetTFRecordDataSet',
        data_source=dict(
            type='ClsSourceImageNetTFRecord',
            file_pattern='oss://path/to/data/imagenet_tfrecord/train-*',
            cache_path='data',
        ),
        pipeline=train_pipeline,
        label_offset=1),
    val=dict(
        imgs_per_gpu=50,
        type='DaliImageNetTFRecordDataSet',
        data_source=dict(
            type='ClsSourceImageNetTFRecord',
            file_pattern='oss://path/to/data/imagenet_tfrecord/validation-*',
            cache_path='data',
        ),
        pipeline=val_pipeline,
        random_shuffle=False,
        label_offset=1))

eval_config = dict(initial=True, interval=1, gpu_collect=True)
eval_pipelines = [
    dict(
        mode='test',
        data='${data.val}',
        dist_eval=True,
        evaluators=[dict(type='${evaluator1_name}', topk=(1, 5))],
    ),
    dict(
        mode='extract',
        dist_eval=True,
        data='${data.val}',
        evaluators=[
            dict(
                type='${evaluator2_name}',
                topk=(1, 2),
                metric_names=('R@K=1', 'R@K=8'),
                feature_keyword=['neck', 'backbone'])
        ],
    )
]

# additional hooks
custom_hooks = []

# optimizer
optimizer = dict(type='SGD', lr=0.1, momentum=0.9, weight_decay=0.0001)

# learning policy
lr_config = dict(policy='step', step=[30, 60, 90])
checkpoint_config = dict(interval=10)

# runtime settings
total_epochs = 100
load_from = None
resume_from = None

export = dict(export_neck=True)



# Copyright (c) Alibaba, Inc. and its affiliates.
from .transforms import Lighting, RandomAppliedTrans, Solarization

__all__ = ['RandomAppliedTrans', 'Lighting', 'Solarization']

try:
    from .transforms import GaussianBlur
    __all__.extend(['GaussianBlur'])
except:
    pass



# Copyright (c) Alibaba, Inc. and its affiliates.

import argparse
import functools
import json
import logging
import os
import random
from multiprocessing import Pool, cpu_count

import numpy as np
from tqdm import tqdm

from easycv.file import io


def get_prior_task_id(keys):
    """"The task id ends with `check` is the highest priority.
    """
    k_list = []
    check_k_list = []
    verify_k_list = []
    for k in keys:
        if k.startswith('label-'):
            if k.endswith('check'):
                check_k_list.append(k)
            elif k.endswith('verify'):
                verify_k_list.append(k)
            else:
                k_list.append(k)

    if len(check_k_list):
        return check_k_list
    if len(k_list):
        return k_list
    if len(verify_k_list):
        return verify_k_list

    return []


def is_itag_v2(row):
    """
    The keyword of the data source is `picUrl` in v1, but is `source` in v2
    """
    if 'source' in row['data']:
        return True
    return False


def parser_manifest_row_str(row_str, classes):
    row = json.loads(row_str.strip())
    _is_itag_v2 = is_itag_v2(row)

    parse_results = {}

    # check img_url
    img_url = row['data']['source']
    if img_url.startswith(('http://', 'https://')):
        logging.warning(
            'Not support http url, only support `oss://`, skip the sample: %s!'
            % img_url)
        return parse_results

    # check task ids
    if _is_itag_v2:
        task_ids = get_prior_task_id(row.keys())
    else:
        task_ids = [
            row_k for row_k in row.keys() if row_k.startswith('label-')
        ]
    if len(task_ids) > 1:
        raise NotImplementedError('Not support multi label task ids: %s!' %
                                  task_ids)
    if not len(task_ids):
        logging.warning('Not find label task id in sample: %s, skip it!' %
                        img_url)
        return parse_results

    ann_json = row[task_ids[0]]
    if not ann_json:
        return parse_results

    bboxes, gt_labels = [], []
    image_size = None
    for result in ann_json['results']:
        if result['type'] != 'image':
            continue
        if not image_size:
            image_size = (result['width'], result['height'])
        objs_list = result['data']

        for obj in objs_list:
            if _is_itag_v2:
                if obj['type'] != 'image/polygon':
                    logging.warning(
                        'Result type should be `image/polygon`, but get %s, skip object %s in %s'
                        % (obj['type'], obj, img_url))
                    continue
                sort_points = sorted(obj['value'], key=sum)
                (x0, y0, x1, y1) = np.concatenate(
                    (sort_points[0], sort_points[-1]), axis=0)
                bboxes.append([x0, y0, x1, y1])
                class_name = list(obj['labels'].values())
                if len(class_name) > 1:
                    raise ValueError(
                        'Not support multi label, get class name  %s!' %
                        class_name)
                gt_labels.append(classes.index(class_name[0]))
            else:
                if obj['type'] != 'image/rectangleLabel':
                    logging.warning(
                        'result type [%s] in %s is not image/rectangleLabel, skip it!'
                        % (obj['type'], img_url))
                    continue
                value = obj['value']
                x, y, w, h = value['x'], value['y'], value['width'], value[
                    'height']
                bnd = [x, y, x + w, y + h]
                class_name = obj['labels'][0]
                bboxes.append(bnd)
                gt_labels.append(classes.index(class_name))
        break

    if len(bboxes) == 0:
        bboxes = np.zeros((0, 4), dtype=np.float32)

    parse_results['filename'] = img_url
    parse_results['img_size'] = image_size
    parse_results['gt_bboxes'] = bboxes
    parse_results['gt_labels'] = gt_labels

    return parse_results


def parse_args():
    parser = argparse.ArgumentParser(
        description='Convert itag detection label to raw detection label')
    parser.add_argument(
        'itag_label_file',
        help='itag label manifest file path',
        type=str,
        default=None)
    parser.add_argument(
        'raw_label_dir',
        help='dir path to save raw label',
        type=str,
        default=None)
    parser.add_argument(
        'class_path', help='classname file path', type=str, default=None)
    parser.add_argument(
        '--split',
        action='store_true',
        help='Whether or not split dataset to train/val set')
    parser.add_argument(
        '--split_ratio',
        type=float,
        default=0.3,
        help='The ratio allocated to val set')
    args = parser.parse_args()
    return args


def build_sample(source_item, classes, parse_fn):
    """Build sample info from source item.
    Args:
        source_item: item of source iterator
        classes: classes list
        parse_fn: parse function to parse source_item, only accepts two params: source_item and classes
        load_img: load image or not, if true, cache all images in memory at init
    """
    result_dict = parse_fn(source_item, classes)

    return result_dict


def build_samples(iterable, process_fn):
    samples_list = []
    num_processes = int(cpu_count() / 2)
    with Pool(processes=num_processes) as p:
        with tqdm(total=len(iterable), desc='Scanning images') as pbar:
            for _, result_dict in enumerate(
                    p.imap_unordered(process_fn, iterable)):
                if result_dict:
                    samples_list.append(result_dict)
                pbar.update()

    return samples_list


def get_source_iterator(itag_label_file):
    with io.open(itag_label_file, 'r') as f:
        rows = f.read().splitlines()
    return rows


def parse_class_list(class_path):
    with open(class_path, 'r') as f:
        rows = f.read().splitlines()
        return rows


def write_raw_label(f, sample_dict):
    w, h = sample_dict['img_size'][0], sample_dict['img_size'][1]
    for box, label in zip(sample_dict['gt_bboxes'], sample_dict['gt_labels']):
        plt, prb = box[:2], box[2:]
        cx, cy, bw, bh = (plt[0] + prb[0]) / 2 / w, (
            plt[1] + prb[1]) / 2 / h, (prb[0] - plt[0]) / w, (prb[1] -
                                                              plt[1]) / h
        res = [str(label), str(cx), str(cy), str(bw), str(bh)]
        f.write(' '.join(res) + '\n')


def main():
    args = parse_args()

    class_list = parse_class_list(args.class_path)
    raw_label_dir_train = os.path.join(args.raw_label_dir, 'train')
    raw_label_dir_val = os.path.join(args.raw_label_dir, 'val')
    if not os.path.exists(raw_label_dir_train):
        os.makedirs(raw_label_dir_train)
    if args.split and not os.path.exists(raw_label_dir_val):
        os.makedirs(raw_label_dir_val)

    source_iter = get_source_iterator(args.itag_label_file)
    process_fn = functools.partial(
        build_sample,
        parse_fn=parser_manifest_row_str,
        classes=class_list,
    )
    samples_list = build_samples(source_iter, process_fn)
    random.shuffle(samples_list)
    data_size = len(samples_list)
    for idx, samples_dict in enumerate(samples_list):
        img_path = samples_dict['filename']
        img_name = os.path.splitext(os.path.basename(img_path))[0]
        lable_path = None
        if args.split and idx < args.split_ratio * data_size:
            lable_path = os.path.join(raw_label_dir_val, img_name + '.txt')
        else:
            lable_path = os.path.join(raw_label_dir_train, img_name + '.txt')
        assert not os.path.exists(
            lable_path), 'file %s already exists' % lable_path
        if len(samples_dict['gt_labels']) <= 0:
            continue
        with open(lable_path, 'w') as f:
            write_raw_label(f, samples_dict)


if __name__ == '__main__':
    main()



# Copyright 2019 Alibaba Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Linear modules."""

from __future__ import absolute_import, division, print_function
import math

import torch

from easycv.core.sailfish.util import (BiasUniformInitializer,
                                       KaimingUniformInitializer,
                                       ModelParallel, RenormUniformInitializer)
from easycv.framework.errors import ValueError


class Linear(torch.nn.Module):
    r"""Applies a linear transformation to the incoming data.
  """

    def __init__(self,
                 in_features,
                 out_features,
                 bias=True,
                 weight_initializer=None,
                 bias_initializer=None,
                 parallel=None):
        super(Linear, self).__init__()
        if isinstance(parallel, ModelParallel):
            if out_features % parallel.world_size != 0:
                raise ValueError(
                    'out_features must be divided by parallel.world_size')
            self.out_features = out_features // parallel.world_size
        else:
            self.out_features = out_features
        self.in_features = in_features
        self.weight = torch.nn.Parameter(
            torch.Tensor(self.out_features, self.in_features))
        if bias:
            self.bias = torch.nn.Parameter(torch.Tensor(self.out_features))
        else:
            self.register_parameter('bias', None)
        self.weight_initializer = weight_initializer
        if weight_initializer is None:
            self.weight_initializer = KaimingUniformInitializer(math.sqrt(5))
        self.bias_initializer = bias_initializer
        if bias_initializer is None:
            self.bias_initializer = BiasUniformInitializer(self.in_features)
        self.reset_parameters()
        self.parallel = parallel

    def reset_parameters(self):
        r"""Reset parameter."""
        self.weight_initializer(self.weight)
        if self.bias is not None:
            self.bias_initializer(self.bias)

    def forward(self, features):  # pylint: disable=arguments-differ
        features = features.type(dtype=self.weight.dtype)
        return torch.nn.functional.linear(features, self.weight, self.bias)


class ArcFaceLinear(torch.nn.Module):
    r"""Applies a ArcFace transformation to the incoming data.
      See https://arxiv.org/abs/1801.05599 .
  """

    def __init__(
            self,
            in_features,
            out_features,
            margin=0.5,
            scale=64.0,  # See normface https://arxiv.org/abs/1704.06369
            fast_phi=False,
            epsilon=0,
            weight_initializer=None,
            l2_norm=False,
            parallel=None):
        super(ArcFaceLinear, self).__init__()
        if isinstance(parallel, ModelParallel):
            if out_features % parallel.world_size != 0:
                raise ValueError(
                    'out_features must be divided by parallel.world_size')
            self.out_features = out_features // parallel.world_size
        else:
            self.out_features = out_features
        self.in_features = in_features
        self.margin = margin  # Angular margin penalty.
        self.scale = scale  # Radius of hybershpere.
        self.fast_phi = fast_phi
        self.epsilon = epsilon
        self.weight_initializer = weight_initializer
        if weight_initializer is None:
            self.weight_initializer = RenormUniformInitializer()
        self.l2_norm = l2_norm
        self.parallel = parallel

        self.weight = torch.nn.Parameter(
            torch.Tensor(self.out_features, self.in_features))
        self.reset_parameters()

        self._cos_margin = math.cos(margin)
        self._sin_margin = math.sin(margin)
        self._threshold = math.cos(math.pi - margin)
        self._min = math.sin(math.pi - margin) * self.margin

    def reset_parameters(self):
        r"""Reset parameters."""
        self.weight_initializer(self.weight)

    def forward(self, features, target):  # pylint: disable=arguments-differ
        r"""Compute ::math`\phi = \cos(\theta + margin)` and logits."""
        # (N, E) x (E, C) -> (N, C)
        features = features.type(dtype=self.weight.dtype)
        if self.l2_norm:
            features_norm = torch.norm(features, 2, 1, True)
            features = torch.div(features, features_norm)
            weight_norm = torch.norm(self.weight, 2, 0, True)
            weight = torch.div(self.weight, weight_norm)
        else:
            features = torch.nn.functional.normalize(features)
            weight = torch.nn.functional.normalize(self.weight)
        cosine = torch.nn.functional.linear(features, weight)
        cosine = cosine.clamp(-1, 1)  # for numerical stability
        sine = torch.sqrt(1. + self.epsilon - cosine * cosine)
        phi = cosine * self._cos_margin - sine * self._sin_margin
        phi = phi.type(dtype=cosine.dtype)
        if self.fast_phi:
            phi = torch.where(cosine > 0, phi, cosine)
        else:
            phi = torch.where(cosine > self._threshold, phi,
                              cosine - self._min)
        if isinstance(self.parallel, ModelParallel):
            mask = self.parallel.correct_mask(target, cosine)
        else:
            mask = torch.zeros(
                cosine.size(), device=cosine.device, dtype=cosine.dtype)
            mask.scatter_(1, target.view(-1, 1).long(), 1)
        logits = (mask * phi) + ((1.0 - mask) * cosine)
        logits *= self.scale
        return logits



# Copyright (c) Alibaba, Inc. and its affiliates.
import glob
import os

from mmcv.runner import Hook
from mmcv.runner.dist_utils import master_only

from easycv.file import io
from .registry import HOOKS


@HOOKS.register_module
class OSSSyncHook(Hook):
    """
    upload log files and checkpoints to oss when training on pai
    """

    def __init__(self,
                 work_dir,
                 oss_work_dir,
                 interval=1,
                 ckpt_filename_tmpl='epoch_{}.pth',
                 export_ckpt_filename_tmpl='epoch_{}_export.pt',
                 other_file_list=[],
                 iter_interval=None):
        """
        Args:
            work_dir: work_dir in cfg
            oss_work_dir:  oss directory where to upload local files in work_dir
            interval: upload frequency
            ckpt_filename_tmpl: checkpoint filename template
            other_file_list: other file need to be upload to oss
            iter_interval: upload frequency by iter interval, default to be None, means do it with certain assignment
        """
        self.work_dir = work_dir
        self.oss_work_dir = oss_work_dir
        self.interval = interval
        self.ckpt_filename_tmpl = ckpt_filename_tmpl
        self.export_ckpt_filename_tmpl = export_ckpt_filename_tmpl
        self.other_file_list = other_file_list
        self.iter_interval = iter_interval

    def upload_file(self, runner):
        if hasattr(runner, 'file_upload_perepoch'):
            up_load_file_list = runner.file_upload_perepoch + self.other_file_list
        else:
            up_load_file_list = self.other_file_list

        up_load_file_list = list(set(up_load_file_list))
        epoch = runner.epoch + 1

        # try to up load ckpt model
        ckpt_fname = self.ckpt_filename_tmpl.format(epoch)
        local_ckpt = os.path.join(self.work_dir, ckpt_fname)
        oss_ckpt = os.path.join(self.oss_work_dir, ckpt_fname)
        if not os.path.exists(local_ckpt):
            runner.logger.warning(f'{local_ckpt} does not exists, skip upload')
        else:
            runner.logger.info(f'upload {local_ckpt} to {oss_ckpt}')
            io.safe_copy(local_ckpt, oss_ckpt)

        for other_file in up_load_file_list:
            local_files = glob.glob(
                os.path.join(self.work_dir, other_file), recursive=True)
            for local_file in local_files:
                rel_path = os.path.relpath(local_file, self.work_dir)
                oss_file = os.path.join(self.oss_work_dir, rel_path)
                runner.logger.info(f'upload {up_load_file_list}')
                io.safe_copy(local_file, oss_file)

        # local_tf_logs = os.path.join(self.work_dir, 'tf_logs')
        # oss_tf_logs = os.path.join(self.oss_work_dir, 'tf_logs')
        # runner.logger.info(f'upload directory {local_tf_logs} to {oss_tf_logs}')
        # io.copytree(local_tf_logs, oss_tf_logs)

    # we still use oss sdk to upload pth, log, by default iter 1000, which
    @master_only
    def after_train_iter(self, runner):
        # upload checkpoint and tf events
        if self.iter_interval is not None:
            if not self.every_n_inner_iters(runner, self.iter_interval):
                return
            self.upload_file(runner)
        return

    @master_only
    def after_train_epoch(self, runner):
        # upload checkpoint and tf events
        if not self.every_n_epochs(runner, self.interval):
            return
        self.upload_file(runner)

    @master_only
    def after_run(self, runner):
        # upload final log files
        timestamp = runner.timestamp
        upload_files = [
            '{}.log'.format(timestamp),
            '{}.log.json'.format(timestamp),
        ]
        for log_file in upload_files:
            local_log = os.path.join(self.work_dir, log_file)
            if not os.path.exists(local_log):
                runner.logger.warning(
                    f'{local_log} does not exists, skip upload')
                continue
            oss_log = os.path.join(self.oss_work_dir, log_file)
            runner.logger.info(f'upload {local_log} to {oss_log}')
            io.safe_copy(local_log, oss_log)

        # try to upload exported model
        epoch = runner.epoch
        export_ckpt_fname = self.export_ckpt_filename_tmpl.format(epoch)

        # upload all export files
        export_files = glob.glob(
            os.path.join(self.work_dir, '*{}*'.format(export_ckpt_fname)),
            recursive=True)
        for export_file in export_files:
            rel_path = os.path.relpath(export_file, self.work_dir)
            target_oss_path = os.path.join(self.oss_work_dir, rel_path)
            runner.logger.info(f'upload {export_file} to {target_oss_path}')
            io.safe_copy(export_file, target_oss_path)
